package Introspector::Find;
use 5.006;
use strict;
use warnings;
use warnings::register;
our $VERSION = '1.09';
require Exporter;
require Cwd;

our @ISA = qw(Exporter);
our @EXPORT = qw(find finddepth);
our $debug_filter=0;
use strict;

require File::Basename;
require File::Spec;

# Should ideally be my() not our() but local() currently
# refuses to operate on lexicals

our %SLnkSeen;
# what things we seen

our ($wanted_callback, $avoid_nlink, $bydepth, $no_chdir, $follow,
     $follow_skip, $full_check, 
     $follow_children,
     $skip_children,
     $filter_children,
     $check_children,
     $untaint, $untaint_skip, $untaint_pat,
     $pre_process, $post_process, $dangling_symlinks);


sub my_chdir
{
    my $dir=shift;
    warn "CD $dir" if $debug_filter;
    if ($check_children)
    {

	if ($check_children->($dir) )
	{
	    warn "Entering $dir" if $debug_filter;
	    return chdir ($dir);
	}
	else
	{
	    warn "Check Failed $dir" if $debug_filter;
	    return undef;
	}
    }
}

sub contract_name {
    my ($cdir,$fn) = @_;
    return substr($cdir,0,rindex($cdir,'/')) if $fn eq $Introspector::Find::current_dir;
    $cdir = substr($cdir,0,rindex($cdir,'/')+1);
    $fn =~ s|^\./||;
    my $abs_name= $cdir . $fn;
    if (substr($fn,0,3) eq '../') {
	1 while $abs_name =~ s!/[^/]*/\.\./!/!;
    }
    return $abs_name;
}

# return the absolute name of a directory or file
sub contract_name_Mac {
    my ($cdir,$fn) = @_; 
    my $abs_name;
    if ($fn =~ /^(:+)(.*)$/) { # valid pathname starting with a ':'
	my $colon_count = length ($1);
	if ($colon_count == 1) {
	    $abs_name = $cdir . $2;
	    return $abs_name;
	}
	else { 
	    # need to move up the tree, but 
	    # only if it's not a volume name
	    for (my $i=1; $i<$colon_count; $i++) {
		unless ($cdir =~ /^[^:]+:$/) { # volume name
		    $cdir =~ s/[^:]+:$//;
		}
		else {
		    return undef;
		}
	    }
	    $abs_name = $cdir . $2;
	    return $abs_name;
	}
    }
    else {
	# $fn may be a valid path to a directory or file or (dangling)
	# symlink, without a leading ':'
	if ( (-e $fn) || (-l $fn) ) {
	    if ($fn =~ /^[^:]+:/) { # a volume name like DataHD:*
		return $fn; # $fn is already an absolute path
	    }
	    else {
		$abs_name = $cdir . $fn;
		return $abs_name;
	    }
	}
	else { # argh!, $fn is not a valid directory/file 
	    return undef;
	}
    }
}

sub PathCombine($$) {
    my ($Base,$Name) = @_;
    my $AbsName;
    if (substr($Name,0,1) eq '/') {
	$AbsName= $Name;
    }
    else {
	$AbsName= contract_name($Base,$Name);
    }
    # (simple) check for recursion
    my $newlen= length($AbsName);
    if ($newlen <= length($Base)) {
	if (($newlen == length($Base) || substr($Base,$newlen,1) eq '/')
	    && $AbsName eq substr($Base,0,$newlen))
	{
	    return undef;
	}
    }
    return $AbsName;
}
sub Follow_SymLink($) {
    my ($AbsName) = @_;
    my ($NewName,$DEV, $INO);
    ($DEV, $INO)= lstat $AbsName;
    while (-l _) {
	if ($SLnkSeen{$DEV, $INO}++) {
	    if ($follow_skip < 2) {
		die "$AbsName is encountered a second time";
	    }
	    else {
#		warn "$AbsName is encountered a second time";
		return undef;
	    }
	}
	$NewName= PathCombine($AbsName, readlink($AbsName));
	unless(defined $NewName) {
	    if ($follow_skip < 2) {
		die "$AbsName is a recursive symbolic link";
	    }
	    else {
		return undef;
	    }
	}
	else {
	    $AbsName= $NewName;
	}
	($DEV, $INO) = lstat($AbsName);
	return undef unless defined $DEV;  #  dangling symbolic link
    }
    if ($full_check && defined $DEV && $SLnkSeen{$DEV, $INO}++) {
	if ( ($follow_skip < 1) || ((-d _) && ($follow_skip < 2)) ) {
	    die "$AbsName encountered a second time";
	}
	else {
#	    warn "$AbsName encountered a second time";
	    return undef;
	}
    }
    return $AbsName;
}

our($dir, $name, $fullname, $prune);

#sub _find_dir_symlnk($$$);
#		     sub _find_dir($$$);
# check whether or not a scalar variable is tainted
# (code straight from the Camel, 3rd ed., page 561)


sub _find_opt {
    my $wanted = shift;
    die "invalid top directory" unless defined $_[0];
    # This function must local()ize everything because callbacks may
    # call find() or finddepth()
    local %SLnkSeen;
    local ($wanted_callback, $avoid_nlink, $bydepth, $no_chdir, $follow,
	   $follow_skip,
	   $follow_children, 
	   $skip_children,
	   $filter_children,
	   $check_children,
	   $full_check, $untaint, $untaint_skip, $untaint_pat,
	   $pre_process, $post_process, $dangling_symlinks);
    local($dir, $name, $fullname, $prune);
    local *_ = \my $a;
    my $cwd            = $wanted->{bydepth} ? Cwd::fastcwd() : Cwd::getcwd();
    my $cwd_untainted  = $cwd;
    my $check_t_cwd    = 1;
    $wanted_callback   = $wanted->{wanted};
    $bydepth           = $wanted->{bydepth};
    $pre_process       = $wanted->{preprocess};
    $post_process      = $wanted->{postprocess};
    $no_chdir          = $wanted->{no_chdir};
    $full_check        = $wanted->{follow};
    $follow            = $full_check || $wanted->{follow_fast};
    $follow_skip       = $wanted->{follow_skip};
    $follow_children   = $wanted->{follow_children};
    $skip_children     = $wanted->{skip_children};
    $filter_children   = $wanted->{filter_children};
    $check_children    = $wanted->{check_children};
#    warn "check_children $check_children";
    $untaint           = $wanted->{untaint};
    $untaint_pat       = $wanted->{untaint_pattern};
    $untaint_skip      = $wanted->{untaint_skip};
    $dangling_symlinks = $wanted->{dangling_symlinks};
    # for compatibility reasons (find.pl, find2perl)
    local our ($topdir, $topdev, $topino, $topmode, $topnlink);
    # a symbolic link to a directory doesn't increase the link count
    $avoid_nlink      = $follow || $Introspector::Find::dont_use_nlink;
    my ($abs_dir, $Is_Dir);
  Proc_Top_Item:

    # ENTRY POINT
    foreach my $TOP (@_) {
	my $top_item = $TOP;
	$top_item =~ s|/\z|| unless $top_item eq '/';
	($topdev,$topino,$topmode,$topnlink) = $follow ? stat $top_item : lstat $top_item;
	$Is_Dir= 0;

	if ($follow) {
	    {
		if (substr($top_item,0,1) eq '/') {
		    $abs_dir = $top_item;
		}
		elsif ($top_item eq $Introspector::Find::current_dir) {
		    $abs_dir = $cwd;
		}
		else {  # care about any  ../
		    $abs_dir = contract_name("$cwd/",$top_item);
		}
	    }
	    #  DB<5> x $abs_dir0  '/mnt/sda10/mdupont/experiments/introspector/gcc/introspector-tests/test2/nodes/f1/OBJ'


	    warn "Going to follow \"$abs_dir\"\n";

	    $abs_dir= Follow_SymLink($abs_dir);

	    unless (defined $abs_dir) {
		if ($dangling_symlinks) {
		    if (ref $dangling_symlinks eq 'CODE') {
			$dangling_symlinks->($top_item, $cwd);
		    } else {
			warnings::warnif "$top_item is a dangling symbolic link\n";
		      }
		}
		next Proc_Top_Item;
	    }

	    if (-d _) {
		_find_dir_symlnk($wanted, $abs_dir, $top_item);
		$Is_Dir= 1;
	    }
	}
	else { # no follow
	    $topdir = $top_item;
	    unless (defined $topnlink) {
		warnings::warnif "Can't stat $top_item: $!\n";
		  next Proc_Top_Item;
	      }
	    if (-d _) {
		_find_dir($wanted, $top_item, $topnlink);
		$Is_Dir= 1;
	    }
	    else {
		$abs_dir= $top_item;
	    }
	}
	unless ($Is_Dir) {
	    unless (($_,$dir) = File::Basename::fileparse($abs_dir)) {
		($dir,$_) = ('./', $top_item);
	    }
	    $abs_dir = $dir;

	    unless ($no_chdir || my_chdir $abs_dir) {
		warnings::warnif "Couldn't chdir $abs_dir: $!\n";
		  next Proc_Top_Item;
	      }
	    $name = $abs_dir . $_; # $Introspector::Find::name
	    $_ = $name if $no_chdir;
	    { $wanted_callback->() }; # protect against wild "next"
	}
	unless ( $no_chdir ) {

	}
    }
}

# API:
#  $wanted
#  $p_dir :  "parent directory"
#  $nlink :  what came back from the stat
# preconditions:
#  chdir (if not no_chdir) to dir

sub _find_dir($$$) {
    my ($wanted, # callback
	$p_dir, 
	$nlink) = @_;
    my ($CdLvl,$Level) = (0,0);
    my @Stack;
    my @filenames;
    my ($subcount,$sub_nlink);
    my $SE= [];
    my $dir_name= $p_dir;
    my $dir_pref;
    my $dir_rel = $Introspector::Find::current_dir;
    my $tainted = 0;
    my $no_nlink;
    $dir_pref= ( $p_dir eq '/' ? '/' : "$p_dir/" );
    local ($dir, $name, $prune, *DIR);
    unless ( $no_chdir || ($p_dir eq $Introspector::Find::current_dir)) {
	my $udir = $p_dir;

	unless (my_chdir ( $udir)) {
	    warnings::warnif "Can't cd 1 to $udir: $!\n";
	      return;
	  }
    }
    # push the starting directory
    push @Stack,[$CdLvl,$p_dir,$dir_rel,-1]  if  $bydepth;
    while (defined $SE) {
	unless ($bydepth) {
	    $dir= $p_dir; # $Introspector::Find::dir 
	    $name= $dir_name; # $Introspector::Find::name 
	    $_= ($no_chdir ? $dir_name : $dir_rel ); # $_
	    # prune may happen here
	    $prune= 0;
	    { $wanted_callback->() };	# protect against wild "next"
	    next if $prune;
	}
	# change to that directory
	unless ($no_chdir || ($dir_rel eq $Introspector::Find::current_dir)) {
	    my $udir= $dir_rel;

	    unless (chdir ( $udir)) {
		warnings::warnif "Can't cd 3 to (" .
		    ($p_dir ne '/' ? $p_dir : '') . "/) $udir: $!\n";
		  next;
	      }
	    $CdLvl++;
	}
	$dir= $dir_name; # $Introspector::Find::dir 
	# Get the list of files in the current directory.
	unless (opendir DIR, ($no_chdir ? $dir_name : $Introspector::Find::current_dir)) {
	    warnings::warnif "Can't opendir($dir_name): $!\n";
	      next;
	  }
	@filenames = readdir DIR;
	closedir(DIR);
	@filenames = $pre_process->(@filenames) if $pre_process;
	warn "Check1  $dir_name";
	push @Stack,[
		     $CdLvl,
		     $dir_name,
		     "",
		     -2]   if $post_process;
	# default: use whatever was specifid
	# (if $nlink >= 2, and $avoid_nlink == 0, this will switch back)
	$no_nlink = $avoid_nlink;
	# if dir has wrong nlink count, force switch to slower stat method
	$no_nlink = 1 if ($nlink < 2);
	if ($nlink == 2 && !$no_nlink) {
	    # This dir has no subdirectories.
	    for my $FN (@filenames) {
		next if $FN =~ $Introspector::Find::skip_pattern;
		$name = $dir_pref . $FN; # $Introspector::Find::name
		$_ = ($no_chdir ? $name : $FN); # $_
		{ $wanted_callback->() }; # protect against wild "next"
	    }
	}
	else {
	    # This dir has subdirectories.
	    $subcount = $nlink - 2;
	    # HACK: insert directories at this position. so as to preserve
	    # the user pre-processed ordering of files.
	    # EG: directory traversal is in user sorted order, not at random.
	    my $stack_top = @Stack;
	    for my $FN (@filenames) {
		next if $FN =~ $Introspector::Find::skip_pattern;
		if ($subcount > 0 || $no_nlink) {
		    # Seen all the subdirs?
		    # check for directoriness.
		    # stat is faster for a file in the current directory
		    $sub_nlink = (lstat ($no_chdir ? $dir_pref . $FN : $FN))[3];
		    if (-d _) {
			--$subcount;
			# HACK: replace push to preserve dir traversal order
			#push @Stack,[$CdLvl,$dir_name,$FN,$sub_nlink];
			splice @Stack, $stack_top, 0,
			[$CdLvl,$dir_name,$FN,$sub_nlink];
		    }
		    else {
			$name = $dir_pref . $FN; # $Introspector::Find::name
			$_= ($no_chdir ? $name : $FN); # $_
			{ $wanted_callback->() }; # protect against wild "next"
		    }
		}
		else {
		    $name = $dir_pref . $FN; # $Introspector::Find::name
		    $_= ($no_chdir ? $name : $FN); # $_
		    { $wanted_callback->() }; # protect against wild "next"
		}
	    }
	}
    }
    continue {
	while ( defined ($SE = pop @Stack) ) {
	    ($Level, $p_dir, $dir_rel, $nlink) = @$SE;
	    if ($CdLvl > $Level && !$no_chdir) {
		my $tmp;
		$tmp = join('/',('..') x ($CdLvl-$Level));
		die "Can't cd 4 to $dir_name" . $tmp
		    unless my_chdir ($tmp);
		$CdLvl = $Level;
	    }
	    $dir_name = ($p_dir eq '/' ? "/$dir_rel" : "$p_dir/$dir_rel");
	    $dir_pref = "$dir_name/";
	    if ( $nlink == -2 ) {
		$name = $dir = $p_dir; # $Introspector::Find::name / dir
		$_ = $Introspector::Find::current_dir;
		$post_process->();		# End-of-directory processing
	    }
	    elsif ( $nlink < 0 ) {  # must be finddepth, report dirname now
		$name = $dir_name;
		{
		    if ( substr($name,-2) eq '/.' ) {
			substr($name, length($name) == 2 ? -1 : -2) = '';
		    }
		    $dir = $p_dir;
		    $_ = ($no_chdir ? $dir_name : $dir_rel );
		    if ( substr($_,-2) eq '/.' ) {
			substr($_, length($_) == 2 ? -1 : -2) = '';
		    }
		}
		{ $wanted_callback->() }; # protect against wild "next"
	    }
	    else {
		warn "Check2  $p_dir";
		push @Stack,[$CdLvl,$p_dir,$dir_rel,-1]  if  $bydepth;
		last;
	    }
	}
    }
}

# API:
#  $wanted
#  $dir_loc : absolute location of a dir
#  $p_dir   : "parent directory"
# preconditions:
#  chdir (if not no_chdir) to dir

sub _find_dir_symlnk($$$) {
    my ($wanted, $dir_loc, $p_dir) = @_; # $dir_loc is the absolute directory
    my @Stack;
    my @filenames;
    my $new_loc;
    my $updir_loc = $dir_loc; # untainted parent directory
    my $SE = [];
    my $dir_name = $p_dir;
    my $dir_pref;
    my $loc_pref;
    my $dir_rel = $Introspector::Find::current_dir;
    my $byd_flag; # flag for pending stack entry if $bydepth
    my $tainted = 0;
    my $ok = 1;
    {
	$dir_pref = ( $p_dir   eq '/' ? '/' : "$p_dir/" );
	$loc_pref = ( $dir_loc eq '/' ? '/' : "$dir_loc/" );
    }
    local ($dir, $name, $fullname, $prune, *DIR);
    unless ($no_chdir) {

	$ok = my_chdir($updir_loc) unless ($p_dir eq $Introspector::Find::current_dir);
	unless ($ok) {
	    warnings::warnif "Can't cd 5 to $updir_loc: $!\n";
	      return;
	  }
    }
    # mdupont, 
    # check if we are at the start of the tree
    # then we can restart the stats
    if ($check_children)
    {
#	warn "MYCheck $p_dir";
	if (!$check_children->($p_dir) )
	{
	    warn "Check Failed";
	}
    }
    push @Stack,[$dir_loc,$updir_loc,$p_dir,$dir_rel,-1]  if  $bydepth;
    while (defined $SE) {
	unless ($bydepth) {
	    # change (back) to parent directory (always untainted)
	    unless ($no_chdir) {
		unless (my_chdir $updir_loc) {
		    warnings::warnif "Can't cd 6 to $updir_loc: $!\n";
		      next;
		  }
	    }
	    $dir= $p_dir; # $Introspector::Find::dir
	    $name= $dir_name; # $Introspector::Find::name
	    $_= ($no_chdir ? $dir_name : $dir_rel ); # $_
	    $fullname= $dir_loc; # $Introspector::Find::fullname
	    # prune may happen here
	    $prune= 0;
	    lstat($_); # make sure  file tests with '_' work
	    { $wanted_callback->() }; # protect against wild "next"
	    next if $prune;
	}
	# change to that directory
	unless ($no_chdir || ($dir_rel eq $Introspector::Find::current_dir)) {
	    $updir_loc = $dir_loc;

	    unless (my_chdir $updir_loc) {
		# this is the place where we want to filter
		#warnings::warnif "Can't cd mike 7 to $updir_loc: $!\n";
		  next;
	      }
	}
	$dir = $dir_name; # $Introspector::Find::dir
	# Get the list of files in the current directory.
	unless (opendir DIR, ($no_chdir ? $dir_loc : $Introspector::Find::current_dir)) {
	    warnings::warnif "Can't opendir($dir_loc): $!\n";
	      next;
	  }
	@filenames = readdir DIR;
	closedir(DIR);
	for my $FN (@filenames) {
	    next if $FN =~ $Introspector::Find::skip_pattern;

	    next unless $filter_children->($FN,$loc_pref);

	    # follow symbolic links / do an lstat
	    $new_loc = Follow_SymLink($loc_pref.$FN);

	    warn "New Loc :$new_loc" if $debug_filter;

	    # ignore if invalid symlink
	    unless (defined $new_loc) {
		if ($dangling_symlinks) {
		    if (ref $dangling_symlinks eq 'CODE') {
			$dangling_symlinks->($FN, $dir_pref);
		    } else {
			warnings::warnif "$dir_pref$FN is a dangling symbolic link\n";
		      }
		}
		$fullname = undef;
		$name = $dir_pref . $FN;
		$_ = ($no_chdir ? $name : $FN);
		{ $wanted_callback->() };
		next;
	    }
	    if (-d _) {
		# mdupont
		if ($check_children)
		{
#		    warn "CHECK1 $_";
		    if ($check_children->($dir_name,$FN))
		    {
#			warn "Faile3";
		    }
		}
#		warn "Check3  $dir_name";
		push @Stack,[$new_loc,$updir_loc,$dir_name,$FN,1];
	    }
	    else {
		$fullname = $new_loc; # $Introspector::Find::fullname 
		$name = $dir_pref . $FN; # $Introspector::Find::name
		$_ = ($no_chdir ? $name : $FN); # $_
		{ $wanted_callback->() }; # protect against wild "next"
	    }
	}
    }
    continue {
	while (defined($SE = pop @Stack)) {
	    ($dir_loc, $updir_loc, $p_dir, $dir_rel, $byd_flag) = @$SE;
	    
	    {
		$dir_name = ($p_dir eq '/' ? "/$dir_rel" : "$p_dir/$dir_rel");
		$dir_pref = "$dir_name/";
		$loc_pref = "$dir_loc/";
	    }
	    if ( $byd_flag < 0 ) {  # must be finddepth, report dirname now
		unless ($no_chdir || ($dir_rel eq $Introspector::Find::current_dir)) {
		    unless (my_chdir $updir_loc) { # $updir_loc (parent dir) is always untainted 
			warnings::warnif "Can't cd 8 to $updir_loc: $!\n";
			  next;
		      }
		}
		$fullname = $dir_loc; # $Introspector::Find::fullname
		$name = $dir_name; # $Introspector::Find::name
		
		{
		    if ( substr($name,-2) eq '/.' ) {
			substr($name, length($name) == 2 ? -1 : -2) = ''; # $Introspector::Find::name
		    }
		    $dir = $p_dir; # $Introspector::Find::dir
		    $_ = ($no_chdir ? $dir_name : $dir_rel); # $_
		    if ( substr($_,-2) eq '/.' ) {
			substr($_, length($_) == 2 ? -1 : -2) = '';
		    }
		}
		lstat($_); # make sure file tests with '_' work
		{ $wanted_callback->() }; # protect against wild "next"
	    }
	    else {
#		warn "Check5  $p_dir";
		push @Stack,[$dir_loc, $updir_loc, $p_dir, $dir_rel,-1]  if  $bydepth;
		last;
	    }
	}
    }
}


sub wrap_wanted {
    my $wanted = shift;
    if ( ref($wanted) eq 'HASH' ) {
	if ( $wanted->{follow} || $wanted->{follow_fast}) {
	    $wanted->{follow_skip} = 1 unless defined $wanted->{follow_skip};
	}
	if ( $wanted->{untaint} ) {
	    $wanted->{untaint_pattern} = $Introspector::Find::untaint_pattern  
		unless defined $wanted->{untaint_pattern};
	    $wanted->{untaint_skip} = 0 unless defined $wanted->{untaint_skip};
	}
	return $wanted;
    }
    else {
	return { wanted => $wanted };
    }
}

sub find {
    my $wanted = shift;
    _find_opt(wrap_wanted($wanted), @_);
}

sub finddepth {
    my $wanted = wrap_wanted(shift);
    $wanted->{bydepth} = 1;
    _find_opt($wanted, @_);
}

# default
$Introspector::Find::skip_pattern    = qr/^\.{1,2}\z/;
$Introspector::Find::untaint_pattern = qr|^([-+@\w./]+)$|;
# These are hard-coded for now, but may move to hint files.
# this _should_ work properly on all platforms
# where Introspector::Find can be expected to work
$Introspector::Find::current_dir = File::Spec->curdir || '.';
$Introspector::Find::dont_use_nlink = 1
    if $^O eq 'os2' || $^O eq 'dos' || $^O eq 'amigaos' || $^O eq 'MSWin32' ||
    $^O eq 'interix' || $^O eq 'cygwin' || $^O eq 'epoc' || $^O eq 'qnx' ||
    $^O eq 'nto';
# Set dont_use_nlink in your hint file if your system's stat doesn't
# report the number of links in a directory as an indication
# of the number of files.
# See, e.g. hints/machten.sh for MachTen 2.2.
unless ($Introspector::Find::dont_use_nlink) {
    require Config;
    $Introspector::Find::dont_use_nlink = 1 if ($Config::Config{'dont_use_nlink'});
}




1;
