#include "config.h"
#include "system.h"
#include "coretypes.h"
#include "tm.h"
#include "tree.h"
#include "introspector_forwards.h"
#include "introspector_stdlib.h"
#include "introspector.h"

static int debug=0;
// find to add 
//slot = htab_find_slot (scalar_evolution_info, &tmp, INSERT);

void 
introspector_predicate_set_domain(introspector_predicate_p field,introspector_type_p type)
{
  field->domain=type;
}

void 
introspector_predicate_set_range(introspector_predicate_p field,introspector_type_p type)
{
  field->range=type;
}

void 
introspector_predicate_add_subject(introspector_predicate_p field,librdf_node* pobject)
{
  introspector_type_add_instance(field->domain,pobject);
}

void 
introspector_predicate_add_object(introspector_predicate_p field,librdf_node* pobject)
{
  introspector_type_add_instance(field->range,pobject);
}


static hashval_t 
introspector_string_hash (librdf_node* n) {
  return (*htab_hash_string) (n->str);
}

static int 
introspector_string_eq (const void *s1_p, const void *s2_p)
{
  librdf_node *s1 = (const char *) s1_p;
  librdf_node *s2 = (const char *) s2_p;
  return strcmp (s1->str, s2->str) == 0;
}

static void 
del_introspector (void *e)
{
  //  free (e);
}

introspector_type_p 
create_introspector_type(
			 const char * name, 
			 introspector_type_p obj
			 )
{
  obj->name=name;
  obj->count=0;
  obj->instances= htab_create (1000, 
			       introspector_string_hash, 
			       introspector_string_eq, 
			       del_introspector);
  return obj;
}
// for each subject/object that is found, add it to the hash
void introspector_type_add_instance(introspector_type_p type,librdf_node* pobject)
{
  //  htab_find ()
//  info = htab_find (cache, &pattern);
  librdf_node **slot;
  librdf_node *node;
  type->count++;
  slot = (librdf_node **)htab_find_slot (type->instances,pobject,INSERT);
  node=*slot;
  if (node != 0)
    {
      node->usage++;
      if(debug)
	printf ("Used %p->%p->,%s:%ld:%ld\n",slot,node,node->str,node->id,node->usage);
    }
  else
    {
      *slot = pobject;
      node=*slot;
      if(debug)
	printf ("Added %p->%p->,%s\n",slot,node,node->str);
    }
}

void 
introspector_type_delete(introspector_type_p type)
{
  introspector_type_report(type);
  htab_delete (type->instances);
}

void 
introspector_predicate_delete(introspector_predicate_p type)
{
}

/* Called via htab_traverse.  Output alloc_pool descriptor pointed out by SLOT
   and update statistics.  */
static int
print_type (void **slot, void *b)
{
  librdf_node *n = (librdf_node*) *slot;
  b=0;
  //  introspector_type_p t = (introspector_type_p) b;

  if (debug)
    printf ("\tPointer:%p %p:",slot,n);
  if (debug)
    printf ("\tVisiting %s %ld\n",n->str,n->usage);
  return 1;
}

void 
introspector_type_report(introspector_type_p type)
{
  if (type->count>0)
    {
      if (debug)
	printf ("Begining of visiting type %s with count %ld\n",type->name,type->count);
    }
  htab_traverse (type->instances, print_type, type);
}


