#include "config.h"
#include "system.h"
#include "coretypes.h"
#include "tm.h"
#include "tree.h"
#include "introspector_forwards.h"
#include "splay-tree.h"
#include "diagnostic.h"
#include "toplev.h"
#include "tree-dump.h"
#include "tree-pass.h"
#include "langhooks.h"
#include "tree-iterator.h"

#include "introspector.h"
#include "introspector_inline.h"
#include "tree-dump-data.h"

#include "errno.h"
//#include "splay-tree.h"
//#include "tree-dump.h"
static int debug=0;

void intrspctr_start_file(introspector_dump_info_p di)
{
  di->function_name[0] = 0; /* there is no current function */
}

void intrspctr_start_function(introspector_dump_info_p di, const char *function_name)
{
  strcpy(di->function_name, function_name); /* there is no current function */
}

void intrspctr_init_model(introspector_dump_info_p di)
{

  /*
    to do : read this from the di
  */
  strcpy(di->librdf_type,"hashes");
  strcpy(di->librdf_storagename,"gccdu");

#ifdef __RDF_BDB_
     strcpy(di->librdf_options,"hash-type='bdb',dir='.',write='yes',new='yes'");
#else
     strcpy(di->librdf_options,"hash-type='memory',dir='.',write='yes',new='yes'");
#endif

  // RDF!
  di->pworld=librdf_new_world();
  librdf_world_open(di->pworld);

  di->pstorage=librdf_new_storage(di->pworld, 
				  di->librdf_type,
				  di->librdf_storagename,
				  di->librdf_options
				  /*
				    new means to overwrite
				   */
				 );

  di->debug =0; // default the debug to 0

  di->pmodel=librdf_new_model(di->pworld, di->pstorage, NULL);

  // the basis for gcc specfic types
  // http://purl.oclc.org/NET/introspector/2002/11/24/logic/
  di->plogic_boolean    =librdf_new_uri(di->pworld,":boolean");
  //http://purl.oclc.org/NET/introspector/2002/11/24/gcc/
  di->pgcc_node_types    =librdf_new_uri(di->pworld,"node_types:");

  //http://purl.oclc.org/NET/introspector/2002/11/24/gcc/
  di->pgcc_node_modifiers=librdf_new_uri(di->pworld,"node_modifiers:");
  //http://purl.oclc.org/NET/introspector/2002/11/24/gcc/
  di->pgcc_node_fields    =librdf_new_uri(di->pworld,"node_fields:");

  // this will change to be something more usefull!
  di->pcurrent_uri         =librdf_new_uri(di->pworld,"#");

  di->pcurrent_functions =  di->pcurrent_addresses=
    di->pcurrent_integers  =    di->pcurrent_strings =
    di->pcurrent_files = di->pcurrent_functions =  di->pcurrent_uri;

  // the current subject, file and function
  di->psubject  = NULL;
  if (debug)
    printf ("reset the subject for %p\n",di);
  di->pgccfield_treecoderef    =INTROSPECTOR_FIELD(tree_code_ref);
  di->pgccfield_filename =INTROSPECTOR_FIELD(filename);
  di->pgccfield_linenumber =INTROSPECTOR_FIELD(linenumber);
  di->pgccfield_treecode =INTROSPECTOR_FIELD(tree_code);
  di->pgccfield_address =INTROSPECTOR_FIELD(address);
  di->pgccfield_string =INTROSPECTOR_FIELD(string);
  di->pgccfield_integer =INTROSPECTOR_FIELD(integer);
  di->pgccfield_modifier =INTROSPECTOR_FIELD(modifier);
}

/* Add T to the end of the queue of nodes to dump.  Returns the index
   assigned to T.  */

void intrspctr_add_statement (di, ppredicate, pobject)
     introspector_dump_info_p di;
     introspector_predicate_p ppredicate;
     librdf_node* pobject;
{
  librdf_statement* pstatement=NULL;
  pstatement=librdf_new_statement(di->pworld);    
  librdf_statement_set_subject(pstatement,di->psubject); // the current object
  if (debug)
    printf ("set the subject of stmt %x to be %x %s for %x\n",pstatement,di->psubject,di->psubject->str,di);

  librdf_statement_set_predicate(pstatement,ppredicate); // the uri passed
  librdf_statement_set_object(pstatement,pobject);       // the object passed

  librdf_model_add_statement(di->pmodel, pstatement);        // book' em

}

void intrspctr_add_statement_typed_string (di, ppredicate, pobject_base, pstring)
     introspector_dump_info_p di;
     introspector_predicate_p ppredicate;
     librdf_uri* pobject_base;
     const char * pstring;
{
  librdf_node* pobject= NULL;
  pobject=  librdf_new_node_from_uri_local_name(di->pworld, 
						pobject_base, 
						pstring);
  intrspctr_add_statement(di, 
			  ppredicate,
			  pobject	  
			  );
}

void intrspctr_add_statement_literal_string (di, ppredicate, pstring)
     introspector_dump_info_p di;
     introspector_predicate_p ppredicate;
     const char * pstring;
{
  librdf_node* pobject;
  pobject=  librdf_new_node_from_literal/*TODO:_escaped*/(di->pworld, 
					 pstring,
					 NULL,
					 0); // create a new node
   intrspctr_add_statement(di, 
			  ppredicate,
			  pobject	  
			  );
  }

librdf_node* 
intrspctr_get_field_predicate (di, pfield_name)
     introspector_dump_info_p di;
     const char * pfield_name;
{
  // TODO, we should cache these, and look it up, do we have this field created?
  return librdf_new_node_from_uri_local_name(di->pworld, 
					     di->pgcc_node_fields, 
					     pfield_name);  
}

void intrspctr_set_node_type ( 
    introspector_dump_info_p di,
    introspector_predicate_p   ppredicate,
    tree t)
{
   const char * code_name; 
   enum tree_index id =TREE_CODE (t);
   code_name = tree_code_name[id]; 

  // a second statement to use the predicate field
  intrspctr_add_statement_typed_string  (di,   
					 ppredicate,					 
					 di->pgcc_node_types,//prefix
					 code_name   // ending
					 );
  introspector_node_type_add_object(
				    id,
				    di->psubject
				    );


}

void intrspctr_node_ref( di, ppredicate, index )
     introspector_dump_info_p di;
     introspector_predicate_p  ppredicate;
     unsigned int  index;
{
  char buffer [INTRSPCTR_BUFFER_MAX];
  sprintf (buffer,"id-%u", index); // format the string
  
  /*
    TODO : take the type of the node that is available 
      code_name = tree_code_name[(int) TREE_CODE (t)];
      (emitted by intrspctr_set_node_type)
      and build that into the uri, so we append for each type of
      node, a different base uri. We could store them in an array of uri bases.
  */
 intrspctr_add_statement_typed_string  (
						di,    
						ppredicate,
						di->pcurrent_uri, // the current uri
						buffer           // the string
						);
}

void intrspctr_open_ipc(
    introspector_dump_info_p di,
    enum tree_dump_index phase,
    const char * pfunction_name
    )
{

  // dump_base_name is a global from toplev.c
  char *command="cat";
 
  //  command = concat ("introspect_gcc.pl --file-name=", dump_base_name, NULL); // append the parameters
//  command = concat (acommand,"cat",NULL);

/*
  command = concat ("perl -MIntrospector::GCC -e'Introspector::GCC::stream(\"",
		    dump_base_name, NULL); // append the parameters

  command = concat (command, "\",\"" , NULL); // space between parameters
  command = concat (command, dump_files[phase].suffix, NULL);

  if (pfunction_name)
    {

      command = concat (command, "\",\"" , NULL); // space between parameters
      command = concat (command, pfunction_name, NULL);
    }

  command = concat (command, "\");\'", NULL);
*/
  if (debug)
    printf ("\ngoing to open dumper `%s'\n", command);
  
  // dump_files[phase].state < 0 ? "w" :todo 
  di->stream = popen (command,  "w");

  if (!di->stream)
      error ("could not open dump file `%s'", command);
  else
  {
    //      fprintf (di->stream, "stream is opened\n");
    //      printf ("\nstream is open `%s'\n", command);
//      printf ("\nstream is open `%s'\n", command);
      dump_files[phase].state = 1;
  }
//  free (command);    
}


void intrspctr_assert_field_true(di, pfield) /* old*/
     introspector_dump_info_p di;
     introspector_predicate_p pfield;
{
  intrspctr_add_statement_typed_string(di, 
				       pfield,
				       di->plogic_boolean ,  // a gcc modififer
				       "true"               // with the value of this string
				       );
}

void intrspctr_dump_filename(di, filename,linenumber)
     introspector_dump_info_p di;
     const char *filename;     
     unsigned int linenumber;
{
  char linebuffer[INTRSPCTR_BUFFER_MAX];
  char filebuffer[INTRSPCTR_BUFFER_MAX];
  char * pc = linebuffer;
  //  sprintf(linebuffer,"%u",linenumber);

  if (strcmp(filename,"<internal>")==0) //<internal>
  {
    sprintf(filebuffer,"filename-gcc-internal");
  }
  else
  {
    sprintf(filebuffer,"filename:%s:%u",filename,linenumber);
  }
  // TODO: localize the filenames and extract the URI for them
  intrspctr_add_statement_typed_string(
				       di,
				       di->pgccfield_filename,
				       di->pcurrent_files,   // this is very important, where the files are located
				       filebuffer
				       );
}

/*
  this routine sets the current node id in the subject
*/
void dump_index_rdf_main (
    introspector_dump_info_p di,
    unsigned int index)
{
  char buffer [INTRSPCTR_BUFFER_MAX];  
  sprintf (buffer,"id-%u", index); // mdupont introspector
  di->psubject=       librdf_new_node_from_literal(di->pworld,buffer,"",0);
  if (debug) 
    printf ("set the current subject to be %x %s for %x\n",di->psubject,di->psubject->str,di);
}

/*
  mdupont:
  this is for dumping function bodies
*/
introspector_dump_info_p introspector_dump_begin_function (
    introspector_dump_info_p di,
    enum tree_dump_index phase,
    tree f)
{
  char buffer[INTRSPCTR_BUFFER_MAX] = ""; // for storing the objects buffer  
  //TODO:dump_info_init(di,NULL);

//  if (!dump_files[phase].state)
//    return NULL;

  /*
	       decl_as_string (fn, TFF_DECL_SPECIFIERS));
      fprintf (stream, " (%s)\n",
	       decl_as_string (DECL_ASSEMBLER_NAME (fn), 0));

*/
  if (DECL_NAME (f))
    {
      if (DECL_NAME (f))
	{
	  sprintf (buffer, ".%s",IDENTIFIER_POINTER (DECL_NAME (f)));
	}
      else
	{
	  sprintf (buffer, ".unknown_function_%p", (void*)f);
	}
    }
  else
    {
      sprintf (buffer, ".unknown_function_%p",(void*) f);
    }
  /*
    dump_base_name
  */

  intrspctr_start_function(di,buffer);
  intrspctr_init_model(di);
  intrspctr_open_ipc(di,phase,buffer);

 			    
  return di;
}

void introspector_dump_end (introspector_dump_info_p di)
{
  // later try with "application/rdf+xml"
  
  if (di->pserializer)
    {
      if (di->stream)
	{
	  
	  librdf_serializer_serialize_model(di->pserializer, di->stream, NULL, di->pmodel);
	  librdf_free_serializer(di->pserializer);
	}
    }

  //  intrpsctr_free(di);
  //  librdf_free_uri(uri);
  librdf_free_uri(di->pgcc_node_types);
  librdf_free_node(di->pgccfield_treecoderef);
  librdf_free_uri(di->pgcc_node_modifiers);
  librdf_free_uri(di->pgcc_node_fields);
  librdf_free_uri(di->pcurrent_uri);

  librdf_free_model(di->pmodel);
  librdf_free_storage(di->pstorage);
  librdf_free_world(di->pworld);

}

/* Finish a tree dump for PHASE. STREAM is the stream created by
   dump_begin.  */
/*
void
introspector_dump_end (enum tree_dump_index phase , FILE *stream)
{
  fclose (stream);
}
*/

void
introspector_dump_pointer (introspector_dump_info_p di, introspector_predicate_p field, void *ptr)
{
  //introspector_predicate_t pfield;
  // introspector replaces this function entirely
    //  char buffer[INTRSPCTR_BUFFER_MAX];
    // sprintf (buffer, "%lx", (long) ptr); // mdupont introspector
/*   intrspctr_add_statement_typed_string( */
/* 		di,		 */
/* 		di->pgccfield_address, */
/* 		di->pcurrent_addresses, */
/* 		buffer */
/* 		);  */

  //add a second statement just to store the name 
/*   pfield = intrspctr_get_field_predicate(di,field); */
/*   intrspctr_add_statement_literal_string( */
/* 		di, */
/* 		pfield, */
/* 		buffer */
/* 		); */

}

/* Dump integer I using FIELD to identify it.  */

void
introspector_dump_int (introspector_dump_info_p di, introspector_predicate_p field, int i)
{
  //introspector_predicate_t pfield;
  // introspector replaces this function totally
  char buffer[INTRSPCTR_BUFFER_MAX];
  sprintf (buffer, "%d", i); // mdupont introspector

  //  pfield = intrspctr_get_field_predicate(di,field);
  intrspctr_add_statement_literal_string(
		di,
		field,
		buffer
		);
}

void
introspector_dump_string (introspector_dump_info_p di, const char *string)
{
  // introspector replaces this function totally
  intrspctr_add_statement_typed_string(
		di,
		di->pgccfield_modifier,    // the type of the current node is a modifier
		di->pgcc_node_modifiers,   // which one? 
		string                     // with the value of this string
		);
}



void intrspctr_set_field_treecode(introspector_dump_info_p di, tree t)
{
  // introspector
  intrspctr_set_node_type(di,
			  di->pgccfield_treecode,
			  t);
}

void intrspctr_add_statement_field_string(introspector_dump_info_p di,const char * string)
{
      intrspctr_add_statement_literal_string(
					   di,
					   di->pgccfield_string,
					   string
					   );
}

void intrspctr_finish(di)
     introspector_dump_info_p di;
{
  //  intrpsctr_serialize(di);
  di->pserializer=librdf_new_serializer(di->pworld, "ntriples", NULL, NULL);

  // later try with "application/rdf+xml"
  
  if (di->pserializer)
    {
      if (di->stream)
	{
	  librdf_serializer_serialize_model(di->pserializer, di->stream, NULL, di->pmodel);
	  librdf_free_serializer(di->pserializer);
	}
    }
  //  intrpsctr_free(di);
  //  librdf_free_uri(uri);
  librdf_free_uri(di->pgcc_node_types);
  librdf_free_node(di->pgccfield_treecoderef);
  librdf_free_uri(di->pgcc_node_modifiers);
  librdf_free_uri(di->pgcc_node_fields);
  librdf_free_uri(di->pcurrent_uri);

  librdf_free_model(di->pmodel);
  librdf_free_storage(di->pstorage);
  librdf_free_world(di->pworld);
}

/* Dump T, and all its children, on STREAM.  */

void
introspector_dump_node (introspector_dump_info_p idi ,tree t, int flags)
{
  struct dump_info di;
  dump_queue_p dq;
  dump_queue_p next_dq;

  /* Initialize the dump-information structure.  */
  di.stream = idi->stream;
  di.index = 0;
  di.column = 0;
  di.queue = 0;
  di.queue_end = 0;
  di.free_list = 0;
  di.flags = flags;
  di.node = t;
  di.nodes = splay_tree_new (splay_tree_compare_pointers, 0,
			     (splay_tree_delete_value_fn) &free);

  idi->di=&di;
  /* Queue up the first node.  */
  queue (&di, t, DUMP_NONE);

  /* Until the queue is empty, keep dumping nodes.  */
  while (di.queue)
    dequeue_and_dump (idi);

  /* Now, clean up.  */
  for (dq = di.free_list; dq; dq = next_dq)
    {
      next_dq = dq->next;
      free (dq);
    }
  splay_tree_delete (di.nodes);
}

void introspector_dump_function (int flags,tree fndecl)
{
    introspector_dump_info di; /* this structure contains all the data about the dumper*/
    introspector_dump_begin_function (&di,TDI_tree_all, fndecl);

//    int flags = 0xffffff;//~(TDF_RAW | TDF_SLIM | TDF_LINENO | TDF_TREE | TDF_RTL | TDF_IPA); //mdupont hacked

    if (di.stream)
    {
	introspector_dump_node (&di,fndecl,flags);
//TODO	introspector_dump_function_to_file (fn, stream, flags);
	introspector_dump_end (&di);
    }
}

/* Begin a tree dump for PHASE. Stores any user supplied flag in
   *FLAG_PTR and returns a stream to write to. If the dump is not
   enabled, returns NULL.
   Multiple calls will reopen and append to the dump file.  */


void introspector_dump_begin (enum tree_dump_index phase, int *flag_ptr)
{
  char *name;
  struct dump_file_info *dfi;
  FILE *stream;

//  if (phase == TDI_none || !dump_enabled_p (phase))
//    return NULL;

  name = get_dump_file_name (phase);
  dfi = get_dump_file_info (phase);
  ///dfi->state < 0 ? "w" : "a"
  printf ("going to open file %s in tree-dump\n", name);

  stream = fopen (name, "w");
  if (!stream)
    error ("could not open dump file %qs: %s in tree-dump", name, strerror (errno));
  else
    dfi->state = 1;
  free (name);

  if (flag_ptr)
    *flag_ptr = dfi->flags;

  return stream;
}

void  introspector_verify_cgraph_node (tree node)
{}

void  introspector_cgraph_finalize_compilation_unit1 ()
{}

void  introspector_cgraph_finalize_compilation_unit2 ()
{}

void  introspector_cgraph_mark_functions_to_output ()
{}

void  introspector_cgraph_remove_unreachable_nodes1 ()
{}

void  introspector_cgraph_remove_unreachable_nodes2 ()
{}

void  introspector_cgraph_remove_unreachable_nodes3()
{}

void introspector_cgraph_decide_recursive_inlining1 (struct cgraph_node *node)
{
/*  if (cgraph_dump_file)
    fprintf (cgraph_dump_file, 
	     "\nPerforming recursive inlining on %s\n",
	     cgraph_node_name (node));
*/
}

void introspector_cgraph_decide_recursive_inlining2 (struct cgraph_node *node)
{
/*    if (cgraph_dump_file)
	fprintf (cgraph_dump_file, 
		 "\n   Inlined %i times, body grown from %i to %i insns\n", n,
		 master_clone->global.insns, node->global.insns);
*/
}
