#ifndef INC_INTROSPECTOR_H
#define INC_INTROSPECTOR_H

#include "introspector_rdf.h"
/*
  Begin Introspector
     Begin Task
        Begin InputFile
           Create Memory Object
	   Create Dump information
        End InputFile
	Emit Task/InputFile info
     End Task	
     Emit Task info
  End Introspector
  
*/
//typedef librdf_node_p introspector_predicate_t;
#ifndef INTROSPECTOR_CONTEXT
#include "introspector_context.h"
#endif
#include "hash.h"

//introspector_predicate_t intrspctr_get_field_predicate (introspector_dump_info_p di,const char *pfield_name);
void introspector_dump_string (introspector_dump_info_p di, const char *string);
void intrspctr_set_field_treecode(introspector_dump_info_p di, tree t);
void intrspctr_dump_filename(introspector_dump_info_p di,const char *filename,unsigned int linenumber);
#define INTRSPCTR_BUFFER_MAX 512
struct introspector_context_s
{
    FILE *stream; // copied from the dump info    
    struct dump_info * di;
    char function_name[INTRSPCTR_BUFFER_MAX]; // a buffer containing the current function name
  /*
    the redland RDF world we are building for the introspector
  */
  librdf_world  * pworld;   // the world
  librdf_storage *pstorage; // the db file, the storage
  librdf_model* pmodel;     // the model in that world, TODO : add one per function bodie
  librdf_serializer* pserializer; // this is how we write xml/rdf
  // for logic in general
  librdf_uri*  plogic_boolean;     // true and false values
  // for all gcc interaction
  librdf_uri*  pgcc_node_types;     // the url of the gcc
  librdf_uri*  pgcc_node_modifiers;     // the url of the gcc
  librdf_uri*  pgcc_node_fields;     // the url of the gcc
  // for all current object
  librdf_uri*  pcurrent_uri; // the uri of the current file
  librdf_uri*  pcurrent_addresses; // the uri of the current file, all memory addresses
  librdf_uri*  pcurrent_integers;  // the uri of the current file, all integers
  librdf_uri*  pcurrent_strings;   // the uri of the current file, all strings
  librdf_uri*  pcurrent_files;     // the uri of the current file, all files
  librdf_uri*  pcurrent_functions; // the uri of the current file, all functions
  // more predicates
  librdf_node* pgccfield_filename;    // the filename predicate of a node
  librdf_node* pgccfield_linenumber;  // the line number of the current node
  librdf_node* pgccfield_treecode;    // the treecode, the type of node
  librdf_node* pgccfield_address;     // the memory address of the object
  librdf_node* pgccfield_modifier;    // modifier
  librdf_node* pgccfield_string;      // if the node has a string value, it is stored here
  librdf_node* pgccfield_integer;     // the integer value if there is one
  librdf_node* pgccfield_treecoderef;    // reference to a type  predicate
  int debug; // the debug output to screen (defaults off)
  // here are helper variables
  librdf_node* psubject;      // the current subject (id), this is used to indicate what node we are dealing with
  char librdf_type[INTRSPCTR_BUFFER_MAX];        // the type of hash
  char librdf_storagename[INTRSPCTR_BUFFER_MAX]; // the name of the storage
  char librdf_options[INTRSPCTR_BUFFER_MAX];     // the options
};
typedef struct introspector_context_s introspector_dump_info;
void intrspctr_add_statement_field_string(introspector_dump_info_p di,const char * string);
void intrspctr_finish(introspector_dump_info_p di);
void intrspctr_start_file(introspector_dump_info_p di);
void intrspctr_start_function(introspector_dump_info_p di, const char *function_name);
void intrspctr_init_model(introspector_dump_info_p di);
void intrspctr_open_ipc(
    introspector_dump_info_p di,
    enum tree_dump_index phase,
    const char * pfunction_name
    );
introspector_dump_info_p dump_begin_function (
    introspector_dump_info_p di,
    enum tree_dump_index phase,
    tree f);
void introspector_dump_end (introspector_dump_info_p di);
//,enum tree_dump_index phase, FILE *stream
void dump_index_rdf_main (introspector_dump_info_p di,unsigned int index);
void introspector_dump_function (int flags,tree fndecl);
void introspector_c_genericize_dump (tree fndecl);
// tree-cfg.h
/* CFG statistics.  */
struct cfg_stats_d
{
  long num_merged_labels;
};
void dequeue_and_dump (introspector_dump_info_p);


introspector_type_p create_introspector_type(const char * name, introspector_type_p obj);

void introspector_type_report(introspector_type_p type);
// for each subject/object that is found, add it to the hash
void introspector_type_add(introspector_type_p type,librdf_node* object);

void introspector_type_delete(introspector_type_p type);

struct introspector_predicate
{
  librdf_node  name;
  introspector_type_p range;
  introspector_type_p domain;
};

struct introspector_type {
  const char * name;
  htab_t instances;
  long  count;
};

void introspector_predicate_delete(introspector_predicate_p field);

void introspector_predicate_set_range(introspector_predicate_p field,introspector_type_p type);
void introspector_predicate_set_domain(introspector_predicate_p field,introspector_type_p type);

void introspector_type_add_instance(introspector_type_p type,librdf_node* pobject);
void introspector_predicate_add_subject(introspector_predicate_p field,librdf_node* pobject);
void introspector_predicate_add_object(introspector_predicate_p field,librdf_node* pobject);

struct introspector_predicate;
void introspector_dump_int (introspector_dump_info_p di, introspector_predicate_p field, int i);
void introspector_dump_pointer (introspector_dump_info_p di, introspector_predicate_p field, void *ptr);
void intrspctr_assert_field_true(introspector_dump_info_p di, introspector_predicate_p field );
void intrspctr_add_statement_typed_string (introspector_dump_info_p di,introspector_predicate_p ppredicate,librdf_uri* pobject_base, const char * pstring);
void intrspctr_add_statement_typed_string (introspector_dump_info_p di,introspector_predicate_p ppredicate,librdf_uri* pobject_base,const char * pstring);
void intrspctr_add_statement_literal_string (introspector_dump_info_p di,introspector_predicate_p ppredicate,const char * pstring);
void intrspctr_node_ref(introspector_dump_info_p di,introspector_predicate_p  ppredicate,unsigned int index);
void intrspctr_set_node_type (introspector_dump_info_p di, introspector_predicate_p ppredicate, tree t);
void intrspctr_add_statement (introspector_dump_info_p di, introspector_predicate_p ppredicate, librdf_node* pobject);

void introspector_node_types_delete();
void introspector_node_types_init();
void introspector_node_type_add_object(
				  enum tree_index id,
				  librdf_node* pobject
				  );

// create directories
int introspector_dir_model_add_statement(librdf_model* model, librdf_statement* statement);
void introspector_dir_model_start(const char * name);
#endif
