#include "introspector_forwards.h"
#include "introspector_stdlib.h"
#include "introspector_gcc.h"
#include "introspector.h"
#include "introspector_rdf.h"
#include "introspector_inline.h"
#include "introspector_pool.h"

//create a directory called nodes
//optionally create a session directory
//create a dir for each function FunctionName

#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>

const char * base_name="nodes";
static char * dump_name="TODO";
const char * separator="/";

static int debug=0;
void introspector_dir_model_start(const char * name)
{ 
  dump_name=name;

  if (debug)
    fprintf(stderr,"start function %s\n",name);

}

void mymkdir(const char *path);
void mymkdir(const char *path)
{
  if (debug)
    printf("mkdir %s\n",path);
  mkdir (path,1000);
}

void catsep(char * buffer, const char * text )
{
  strcat (buffer,text);  
  strcat (buffer,separator);  
}

void catsepmk(char * buffer, const char * text , int mkdir)
{
  catsep(buffer, text );
  if (mkdir)
    mymkdir(buffer); 
}

void base_dir(char * buffer, int mkdir)
{
  catsepmk (buffer,base_name,mkdir);   
  catsepmk (buffer,dump_name,mkdir);
}

int introspector_base_link(char * buffer, int levels)
{
  char cwdbuffer [INTRSPCTR_BUFFER_MAX]="";  
  getcwd (cwdbuffer, sizeof(cwdbuffer));
  strcat(buffer,cwdbuffer);
  strcat(buffer,"/");
  base_dir(buffer,1);  
}

// Link [from --> ../subject]
int introspector_subject_link(char * from,  librdf_statement* statement)
{
  char buffer [INTRSPCTR_BUFFER_MAX]="";
  introspector_base_link(buffer,3);          // create the right number of ..s 
  catsepmk (buffer,statement->predicate->domain->name,1);   
//  strcat (buffer,"S-");                                  // object of the statement
  catsepmk (buffer,statement->subject->str,0);    // add the subject
  symlink(buffer,from);                           // link from to the subject
  if (debug)
    fprintf(stderr,"subj link %s <-- %s\n",buffer,from);
}

/*
obj link     ../../../../nodes/queue_and_dump_type/STRING/FIELD_OBJECT_type.49/ <-- 
             1    2                 3     4    5    6     7
         nodes/queue_and_dump_type/OBJ/id-55/strg/FIELD_OBJECT_type.49
*/
int introspector_object_link(char * from,  librdf_statement* statement,int levels)
{
  char buffer [INTRSPCTR_BUFFER_MAX]="";
  // link into the typed object
  introspector_base_link(buffer, 4);   // the ..s
  catsepmk (buffer,statement->predicate->range->name,1);   
  catsepmk (buffer,statement->object->str,1); 
  symlink(buffer,from);
  if (debug)
    fprintf(stderr,"obj link %s <-- %s\n",buffer,from);
}

// /Domain/Subject/Predicate/Object
/*
introspector_build_basic_subject:nodes/fstat/OBJ/id-1/args/id-4
*/
int introspector_build_basic_subject(char * buffer,librdf_statement* statement){  // the basic object
  base_dir(buffer,1);
  catsepmk (buffer,statement->predicate->domain->name,1);   // domain of the predicate
//  strcat (buffer,"S-");                                  // object of the statement
  catsepmk (buffer,statement->subject->str,1);              // subject, matches the domain
//  strcat (buffer,"S-");                                  // object of the statement
  catsepmk (buffer,statement->predicate->name.str,1);       // name of the predicate
//  strcat (buffer,"S-");                                  // object of the statement
  strcat (buffer,statement->object->str);                   // object of the statement
  if (debug)
    fprintf(stderr,"introspector_build_basic_subject:%s\n",buffer);

}

/*
to collect again, added the predicate in here.
introspector_dir_model_link_obj:nodes/fstat64/FILENAME/filename-built-in/SUBJ-id-15
*/

int introspector_dir_model_link_obj(librdf_statement* statement) // /Range/Object/Subject-->../Subject
{  
  char obj_buffer [INTRSPCTR_BUFFER_MAX]="";
  //create the link to the subject obj, in the object dir...
  base_dir(obj_buffer,1);
  catsepmk (obj_buffer,statement->predicate->range->name,1);   // RANGE
//  strcat (obj_buffer,"O-");                                    // prefix the predicate to denote a reference
  catsepmk (obj_buffer,statement->object->str,1);              // Object
  strcat (obj_buffer,"O-");                                    // prefix the predicate to denote a reference
  catsepmk (obj_buffer,statement->predicate->name.str,1);      // name of the predicate
  strcat (obj_buffer,"O-");                                    // prefix the subject to denote a reference
  strcat (obj_buffer,statement->subject->str);                 // Subject
  introspector_subject_link(obj_buffer,statement);           // link from the object to the subject Link [from --> ../subject]
  if (debug)
    fprintf(stderr,"introspector_dir_model_link_obj:%s\n",obj_buffer);
}

// the entry into this program. one call per statement.
/*
introspector_dir_model_add_statement:nodes/fstat64/OBJ/id-39/filename/filename-stat.h
*/
int introspector_dir_model_add_statement(librdf_model* model, librdf_statement* statement)
{
  model=0;

  // step 1
  char buffer [INTRSPCTR_BUFFER_MAX]="";
  introspector_build_basic_subject(buffer,statement); // build a string Domain/Subject/Predicate/Object
  introspector_object_link(buffer,statement,2);       // create a link from the type [link to ../range/object]

  // step 2 
//  introspector_dir_model_link_obj(statement);   // Range/Object/Subject-->Object
  if (debug)
    fprintf(stderr,"introspector_dir_model_add_statement:%s\n",buffer);
}
