#define NULL 0
extern FILE *popen (__const char *__command, __const char *__modes);


enum tree_code {ERROR_MARK,IDENTIFIER_NODE,TREE_LIST,TREE_VEC,BLOCK,OFFSET_TYPE,ENUMERAL_TYPE,BOOLEAN_TYPE,CHAR_TYPE,INTEGER_TYPE,REAL_TYPE,POINTER_TYPE,REFERENCE_TYPE,COMPLEX_TYPE,VECTOR_TYPE,ARRAY_TYPE,RECORD_TYPE,UNION_TYPE,QUAL_UNION_TYPE,VOID_TYPE,FUNCTION_TYPE,METHOD_TYPE,FILE_TYPE,LANG_TYPE,INTEGER_CST,REAL_CST,COMPLEX_CST,VECTOR_CST,STRING_CST,FUNCTION_DECL,LABEL_DECL,FIELD_DECL,VAR_DECL,CONST_DECL,PARM_DECL,TYPE_DECL,RESULT_DECL,NAMESPACE_DECL,TRANSLATION_UNIT_DECL,COMPONENT_REF,BIT_FIELD_REF,INDIRECT_REF,ALIGN_INDIRECT_REF,MISALIGNED_INDIRECT_REF,ARRAY_REF,ARRAY_RANGE_REF,OBJ_TYPE_REF,EXC_PTR_EXPR,FILTER_EXPR,CONSTRUCTOR,COMPOUND_EXPR,MODIFY_EXPR,INIT_EXPR,TARGET_EXPR,COND_EXPR,VEC_COND_EXPR,BIND_EXPR,CALL_EXPR,WITH_CLEANUP_EXPR,CLEANUP_POINT_EXPR,PLACEHOLDER_EXPR,PLUS_EXPR,MINUS_EXPR,MULT_EXPR,TRUNC_DIV_EXPR,CEIL_DIV_EXPR,FLOOR_DIV_EXPR,ROUND_DIV_EXPR,TRUNC_MOD_EXPR,CEIL_MOD_EXPR,FLOOR_MOD_EXPR,ROUND_MOD_EXPR,RDIV_EXPR,EXACT_DIV_EXPR,FIX_TRUNC_EXPR,FIX_CEIL_EXPR,FIX_FLOOR_EXPR,FIX_ROUND_EXPR,FLOAT_EXPR,NEGATE_EXPR,MIN_EXPR,MAX_EXPR,ABS_EXPR,LSHIFT_EXPR,RSHIFT_EXPR,LROTATE_EXPR,RROTATE_EXPR,BIT_IOR_EXPR,BIT_XOR_EXPR,BIT_AND_EXPR,BIT_NOT_EXPR,TRUTH_ANDIF_EXPR,TRUTH_ORIF_EXPR,TRUTH_AND_EXPR,TRUTH_OR_EXPR,TRUTH_XOR_EXPR,TRUTH_NOT_EXPR,LT_EXPR,LE_EXPR,GT_EXPR,GE_EXPR,EQ_EXPR,NE_EXPR,UNORDERED_EXPR,ORDERED_EXPR,UNLT_EXPR,UNLE_EXPR,UNGT_EXPR,UNGE_EXPR,UNEQ_EXPR,LTGT_EXPR,RANGE_EXPR,CONVERT_EXPR,NOP_EXPR,NON_LVALUE_EXPR,VIEW_CONVERT_EXPR,SAVE_EXPR,ADDR_EXPR,FDESC_EXPR,COMPLEX_EXPR,CONJ_EXPR,REALPART_EXPR,IMAGPART_EXPR,PREDECREMENT_EXPR,PREINCREMENT_EXPR,POSTDECREMENT_EXPR,POSTINCREMENT_EXPR,VA_ARG_EXPR,TRY_CATCH_EXPR,TRY_FINALLY_EXPR,DECL_EXPR,LABEL_EXPR,GOTO_EXPR,RETURN_EXPR,EXIT_EXPR,LOOP_EXPR,SWITCH_EXPR,CASE_LABEL_EXPR,RESX_EXPR,ASM_EXPR,SSA_NAME,PHI_NODE,CATCH_EXPR,EH_FILTER_EXPR,SCEV_KNOWN,SCEV_NOT_KNOWN,POLYNOMIAL_CHREC,STATEMENT_LIST,VALUE_HANDLE,TREE_BINFO,WITH_SIZE_EXPR,REALIGN_LOAD_EXPR, LAST_AND_UNUSED_TREE_CODE};

//
union tree_node;
typedef union tree_node *tree;


/// basic gcc structures used by the introspector

struct dump_queue;
struct splay_tree_node_s;
struct splay_tree_s;

typedef struct splay_tree_s *splay_tree;
typedef struct splay_tree_node_s *splay_tree_node;

typedef struct dump_queue
{
  splay_tree_node node;
  struct dump_queue *next;
} *dump_queue_p;

struct dump_info
{
  FILE *stream;
  tree node;
  int flags;
  unsigned int index;
  unsigned int column;
  dump_queue_p queue;
  dump_queue_p queue_end;
  dump_queue_p free_list;
  splay_tree nodes;
};
typedef struct dump_info *dump_info_p;




union tree_node;
union tree_ann_d;

struct tree_common
{
  tree chain;
  tree type;
  union tree_ann_d *ann;
  __extension__ enum tree_code code : 8;
  unsigned side_effects_flag : 1;
  unsigned constant_flag : 1;
  unsigned addressable_flag : 1;
  unsigned volatile_flag : 1;
  unsigned readonly_flag : 1;
  unsigned unsigned_flag : 1;
  unsigned asm_written_flag: 1;
  unsigned nowarning_flag : 1;
  unsigned used_flag : 1;
  unsigned nothrow_flag : 1;
  unsigned static_flag : 1;
  unsigned public_flag : 1;
  unsigned private_flag : 1;
  unsigned protected_flag : 1;
  unsigned deprecated_flag : 1;
  unsigned invariant_flag : 1;
  unsigned lang_flag_0 : 1;
  unsigned lang_flag_1 : 1;
  unsigned lang_flag_2 : 1;
  unsigned lang_flag_3 : 1;
  unsigned lang_flag_4 : 1;
  unsigned lang_flag_5 : 1;
  unsigned lang_flag_6 : 1;
  unsigned visited : 1;
};



struct tree_int_cst
{
  struct tree_common common;

  struct tree_int_cst_lowhi {
    unsigned long low;
    long high;
  } int_cst;
};




struct real_value;




struct tree_real_cst
{
  struct tree_common common;
  struct real_value * real_cst_ptr;
};

struct tree_string
{
  struct tree_common common;
  int length;
  char str[1];
};


struct tree_complex
{
  struct tree_common common;
  tree real;
  tree imag;
};

struct tree_vector
{
  struct tree_common common;
  tree elements;
};


typedef struct ht_identifier ht_identifier;
struct ht_identifier
{
  const unsigned char *str;
  unsigned int len;
  unsigned int hash_value;
};

struct tree_identifier
{
  struct tree_common common;
  struct ht_identifier id;
};


struct tree_list
{
  struct tree_common common;
  tree purpose;
  tree value;
};

struct tree_vec
{
  struct tree_common common;
  int length;
  tree a[1];
};

struct location_s
{
  const char *file;
  int line;
};

typedef struct location_s expanded_location;
typedef struct location_s location_t;
typedef location_t *source_locus;

struct tree_exp
{
  struct tree_common common;
  source_locus locus;
  int complexity;
  tree block;
  tree    operands[1];
};

struct rtx_def;
typedef struct rtx_def *rtx;


enum machine_mode
{
  VOIDmode,
  BLKmode,
  CCmode,
  CCGCmode,
  CCGOCmode,
  CCNOmode,
  CCZmode,
  CCFPmode,
  CCFPUmode,
  BImode,
  QImode,
  HImode,
  SImode,
  DImode,
  TImode,
  SFmode,
  DFmode,
  XFmode,
  TFmode,
  CQImode,
  CHImode,
  CSImode,
  CDImode,
  CTImode,
  SCmode,
  DCmode,
  XCmode,
  TCmode,
  V4QImode,
  V2HImode,
  V8QImode,
  V4HImode,
  V2SImode,
  V16QImode,
  V8HImode,
  V4SImode,
  V2DImode,
  V32QImode,
  V16HImode,
  V8SImode,
  V4DImode,
  V2SFmode,
  V4SFmode,
  V2DFmode,
  V8SFmode,
  V4DFmode,
  MAX_MACHINE_MODE,

  MIN_MODE_RANDOM = VOIDmode,
  MAX_MODE_RANDOM = BLKmode,

  MIN_MODE_CC = CCmode,
  MAX_MODE_CC = CCFPUmode,

  MIN_MODE_INT = QImode,
  MAX_MODE_INT = TImode,

  MIN_MODE_PARTIAL_INT = VOIDmode,
  MAX_MODE_PARTIAL_INT = VOIDmode,

  MIN_MODE_FLOAT = SFmode,
  MAX_MODE_FLOAT = TFmode,

  MIN_MODE_COMPLEX_INT = CQImode,
  MAX_MODE_COMPLEX_INT = CTImode,

  MIN_MODE_COMPLEX_FLOAT = SCmode,
  MAX_MODE_COMPLEX_FLOAT = TCmode,

  MIN_MODE_VECTOR_INT = V4QImode,
  MAX_MODE_VECTOR_INT = V4DImode,

  MIN_MODE_VECTOR_FLOAT = V2SFmode,
  MAX_MODE_VECTOR_FLOAT = V4DFmode,

  NUM_MACHINE_MODES = MAX_MACHINE_MODE
};


enum built_in_class
{
  NOT_BUILT_IN = 0,
  BUILT_IN_FRONTEND,
  BUILT_IN_MD,
  BUILT_IN_NORMAL
};

enum symbol_visibility
{
  VISIBILITY_DEFAULT,
  VISIBILITY_INTERNAL,
  VISIBILITY_HIDDEN,
  VISIBILITY_PROTECTED
};


// "../../gcc/builtins.def"  "../../gcc/builtins.def"  "../../gcc/tree.h" 2

enum built_in_function
{
BUILT_IN_ACOS,
BUILT_IN_ACOSF,
BUILT_IN_ACOSH,
BUILT_IN_ACOSHF,
BUILT_IN_ACOSHL,
BUILT_IN_ACOSL,
BUILT_IN_ASIN,
BUILT_IN_ASINF,
BUILT_IN_ASINH,
BUILT_IN_ASINHF,
BUILT_IN_ASINHL,
BUILT_IN_ASINL,
BUILT_IN_ATAN,
BUILT_IN_ATAN2,
BUILT_IN_ATAN2F,
BUILT_IN_ATAN2L,
BUILT_IN_ATANF,
BUILT_IN_ATANH,
BUILT_IN_ATANHF,
BUILT_IN_ATANHL,
BUILT_IN_ATANL,
BUILT_IN_CBRT,
BUILT_IN_CBRTF,
BUILT_IN_CBRTL,
BUILT_IN_CEIL,
BUILT_IN_CEILF,
BUILT_IN_CEILL,
BUILT_IN_COPYSIGN,
BUILT_IN_COPYSIGNF,
BUILT_IN_COPYSIGNL,
BUILT_IN_COS,
BUILT_IN_COSF,
BUILT_IN_COSH,
BUILT_IN_COSHF,
BUILT_IN_COSHL,
BUILT_IN_COSL,
BUILT_IN_DREM,
BUILT_IN_DREMF,
BUILT_IN_DREML,
BUILT_IN_ERF,
BUILT_IN_ERFC,
BUILT_IN_ERFCF,
BUILT_IN_ERFCL,
BUILT_IN_ERFF,
BUILT_IN_ERFL,
BUILT_IN_EXP,
BUILT_IN_EXP10,
BUILT_IN_EXP10F,
BUILT_IN_EXP10L,
BUILT_IN_EXP2,
BUILT_IN_EXP2F,
BUILT_IN_EXP2L,
BUILT_IN_EXPF,
BUILT_IN_EXPL,
BUILT_IN_EXPM1,
BUILT_IN_EXPM1F,
BUILT_IN_EXPM1L,
BUILT_IN_FABS,
BUILT_IN_FABSF,
BUILT_IN_FABSL,
BUILT_IN_FDIM,
BUILT_IN_FDIMF,
BUILT_IN_FDIML,
BUILT_IN_FLOOR,
BUILT_IN_FLOORF,
BUILT_IN_FLOORL,
BUILT_IN_FMA,
BUILT_IN_FMAF,
BUILT_IN_FMAL,
BUILT_IN_FMAX,
BUILT_IN_FMAXF,
BUILT_IN_FMAXL,
BUILT_IN_FMIN,
BUILT_IN_FMINF,
BUILT_IN_FMINL,
BUILT_IN_FMOD,
BUILT_IN_FMODF,
BUILT_IN_FMODL,
BUILT_IN_FREXP,
BUILT_IN_FREXPF,
BUILT_IN_FREXPL,
BUILT_IN_GAMMA,
BUILT_IN_GAMMAF,
BUILT_IN_GAMMAL,
BUILT_IN_HUGE_VAL,
BUILT_IN_HUGE_VALF,
BUILT_IN_HUGE_VALL,
BUILT_IN_HYPOT,
BUILT_IN_HYPOTF,
BUILT_IN_HYPOTL,
BUILT_IN_ILOGB,
BUILT_IN_ILOGBF,
BUILT_IN_ILOGBL,
BUILT_IN_INF,
BUILT_IN_INFF,
BUILT_IN_INFL,
BUILT_IN_J0,
BUILT_IN_J0F,
BUILT_IN_J0L,
BUILT_IN_J1,
BUILT_IN_J1F,
BUILT_IN_J1L,
BUILT_IN_JN,
BUILT_IN_JNF,
BUILT_IN_JNL,
BUILT_IN_LDEXP,
BUILT_IN_LDEXPF,
BUILT_IN_LDEXPL,
BUILT_IN_LGAMMA,
BUILT_IN_LGAMMAF,
BUILT_IN_LGAMMAL,
BUILT_IN_LLRINT,
BUILT_IN_LLRINTF,
BUILT_IN_LLRINTL,
BUILT_IN_LLROUND,
BUILT_IN_LLROUNDF,
BUILT_IN_LLROUNDL,
BUILT_IN_LOG,
BUILT_IN_LOG10,
BUILT_IN_LOG10F,
BUILT_IN_LOG10L,
BUILT_IN_LOG1P,
BUILT_IN_LOG1PF,
BUILT_IN_LOG1PL,
BUILT_IN_LOG2,
BUILT_IN_LOG2F,
BUILT_IN_LOG2L,
BUILT_IN_LOGB,
BUILT_IN_LOGBF,
BUILT_IN_LOGBL,
BUILT_IN_LOGF,
BUILT_IN_LOGL,
BUILT_IN_LRINT,
BUILT_IN_LRINTF,
BUILT_IN_LRINTL,
BUILT_IN_LROUND,
BUILT_IN_LROUNDF,
BUILT_IN_LROUNDL,
BUILT_IN_MODF,
BUILT_IN_MODFF,
BUILT_IN_MODFL,
BUILT_IN_NAN,
BUILT_IN_NANF,
BUILT_IN_NANL,
BUILT_IN_NANS,
BUILT_IN_NANSF,
BUILT_IN_NANSL,
BUILT_IN_NEARBYINT,
BUILT_IN_NEARBYINTF,
BUILT_IN_NEARBYINTL,
BUILT_IN_NEXTAFTER,
BUILT_IN_NEXTAFTERF,
BUILT_IN_NEXTAFTERL,
BUILT_IN_NEXTTOWARD,
BUILT_IN_NEXTTOWARDF,
BUILT_IN_NEXTTOWARDL,
BUILT_IN_POW,
BUILT_IN_POW10,
BUILT_IN_POW10F,
BUILT_IN_POW10L,
BUILT_IN_POWF,
BUILT_IN_POWI,
BUILT_IN_POWIF,
BUILT_IN_POWIL,
BUILT_IN_POWL,
BUILT_IN_REMAINDER,
BUILT_IN_REMAINDERF,
BUILT_IN_REMAINDERL,
BUILT_IN_REMQUO,
BUILT_IN_REMQUOF,
BUILT_IN_REMQUOL,
BUILT_IN_RINT,
BUILT_IN_RINTF,
BUILT_IN_RINTL,
BUILT_IN_ROUND,
BUILT_IN_ROUNDF,
BUILT_IN_ROUNDL,
BUILT_IN_SCALB,
BUILT_IN_SCALBF,
BUILT_IN_SCALBL,
BUILT_IN_SCALBLN,
BUILT_IN_SCALBLNF,
BUILT_IN_SCALBLNL,
BUILT_IN_SCALBN,
BUILT_IN_SCALBNF,
BUILT_IN_SCALBNL,
BUILT_IN_SIGNBIT,
BUILT_IN_SIGNBITF,
BUILT_IN_SIGNBITL,
BUILT_IN_SIGNIFICAND,
BUILT_IN_SIGNIFICANDF,
BUILT_IN_SIGNIFICANDL,
BUILT_IN_SIN,
BUILT_IN_SINCOS,
BUILT_IN_SINCOSF,
BUILT_IN_SINCOSL,
BUILT_IN_SINF,
BUILT_IN_SINH,
BUILT_IN_SINHF,
BUILT_IN_SINHL,
BUILT_IN_SINL,
BUILT_IN_SQRT,
BUILT_IN_SQRTF,
BUILT_IN_SQRTL,
BUILT_IN_TAN,
BUILT_IN_TANF,
BUILT_IN_TANH,
BUILT_IN_TANHF,
BUILT_IN_TANHL,
BUILT_IN_TANL,
BUILT_IN_TGAMMA,
BUILT_IN_TGAMMAF,
BUILT_IN_TGAMMAL,
BUILT_IN_TRUNC,
BUILT_IN_TRUNCF,
BUILT_IN_TRUNCL,
BUILT_IN_Y0,
BUILT_IN_Y0F,
BUILT_IN_Y0L,
BUILT_IN_Y1,
BUILT_IN_Y1F,
BUILT_IN_Y1L,
BUILT_IN_YN,
BUILT_IN_YNF,
BUILT_IN_YNL,
BUILT_IN_CABS,
BUILT_IN_CABSF,
BUILT_IN_CABSL,
BUILT_IN_CACOS,
BUILT_IN_CACOSF,
BUILT_IN_CACOSH,
BUILT_IN_CACOSHF,
BUILT_IN_CACOSHL,
BUILT_IN_CACOSL,
BUILT_IN_CARG,
BUILT_IN_CARGF,
BUILT_IN_CARGL,
BUILT_IN_CASIN,
BUILT_IN_CASINF,
BUILT_IN_CASINH,
BUILT_IN_CASINHF,
BUILT_IN_CASINHL,
BUILT_IN_CASINL,
BUILT_IN_CATAN,
BUILT_IN_CATANF,
BUILT_IN_CATANH,
BUILT_IN_CATANHF,
BUILT_IN_CATANHL,
BUILT_IN_CATANL,
BUILT_IN_CCOS,
BUILT_IN_CCOSF,
BUILT_IN_CCOSH,
BUILT_IN_CCOSHF,
BUILT_IN_CCOSHL,
BUILT_IN_CCOSL,
BUILT_IN_CEXP,
BUILT_IN_CEXPF,
BUILT_IN_CEXPL,
BUILT_IN_CIMAG,
BUILT_IN_CIMAGF,
BUILT_IN_CIMAGL,
BUILT_IN_CONJ,
BUILT_IN_CONJF,
BUILT_IN_CONJL,
BUILT_IN_CPOW,
BUILT_IN_CPOWF,
BUILT_IN_CPOWL,
BUILT_IN_CPROJ,
BUILT_IN_CPROJF,
BUILT_IN_CPROJL,
BUILT_IN_CREAL,
BUILT_IN_CREALF,
BUILT_IN_CREALL,
BUILT_IN_CSIN,
BUILT_IN_CSINF,
BUILT_IN_CSINH,
BUILT_IN_CSINHF,
BUILT_IN_CSINHL,
BUILT_IN_CSINL,
BUILT_IN_CSQRT,
BUILT_IN_CSQRTF,
BUILT_IN_CSQRTL,
BUILT_IN_CTAN,
BUILT_IN_CTANF,
BUILT_IN_CTANH,
BUILT_IN_CTANHF,
BUILT_IN_CTANHL,
BUILT_IN_CTANL,
BUILT_IN_BCMP,
BUILT_IN_BCOPY,
BUILT_IN_BZERO,
BUILT_IN_INDEX,
BUILT_IN_MEMCMP,
BUILT_IN_MEMCPY,
BUILT_IN_MEMMOVE,
BUILT_IN_MEMPCPY,
BUILT_IN_MEMSET,
BUILT_IN_RINDEX,
BUILT_IN_STPCPY,
BUILT_IN_STRCAT,
BUILT_IN_STRCHR,
BUILT_IN_STRCMP,
BUILT_IN_STRCPY,
BUILT_IN_STRCSPN,
BUILT_IN_STRDUP,
BUILT_IN_STRLEN,
BUILT_IN_STRNCAT,
BUILT_IN_STRNCMP,
BUILT_IN_STRNCPY,
BUILT_IN_STRPBRK,
BUILT_IN_STRRCHR,
BUILT_IN_STRSPN,
BUILT_IN_STRSTR,
BUILT_IN_FPRINTF,
BUILT_IN_FPRINTF_UNLOCKED,
BUILT_IN_FPUTC,
BUILT_IN_FPUTC_UNLOCKED,
BUILT_IN_FPUTS,
BUILT_IN_FPUTS_UNLOCKED,
BUILT_IN_FSCANF,
BUILT_IN_FWRITE,
BUILT_IN_FWRITE_UNLOCKED,
BUILT_IN_PRINTF,
BUILT_IN_PRINTF_UNLOCKED,
BUILT_IN_PUTCHAR,
BUILT_IN_PUTCHAR_UNLOCKED,
BUILT_IN_PUTS,
BUILT_IN_PUTS_UNLOCKED,
BUILT_IN_SCANF,
BUILT_IN_SNPRINTF,
BUILT_IN_SPRINTF,
BUILT_IN_SSCANF,
BUILT_IN_VFPRINTF,
BUILT_IN_VFSCANF,
BUILT_IN_VPRINTF,
BUILT_IN_VSCANF,
BUILT_IN_VSNPRINTF,
BUILT_IN_VSPRINTF,
BUILT_IN_VSSCANF,
BUILT_IN_ISALNUM,
BUILT_IN_ISALPHA,
BUILT_IN_ISASCII,
BUILT_IN_ISBLANK,
BUILT_IN_ISCNTRL,
BUILT_IN_ISDIGIT,
BUILT_IN_ISGRAPH,
BUILT_IN_ISLOWER,
BUILT_IN_ISPRINT,
BUILT_IN_ISPUNCT,
BUILT_IN_ISSPACE,
BUILT_IN_ISUPPER,
BUILT_IN_ISXDIGIT,
BUILT_IN_TOASCII,
BUILT_IN_TOLOWER,
BUILT_IN_TOUPPER,
BUILT_IN_ISWALNUM,
BUILT_IN_ISWALPHA,
BUILT_IN_ISWBLANK,
BUILT_IN_ISWCNTRL,
BUILT_IN_ISWDIGIT,
BUILT_IN_ISWGRAPH,
BUILT_IN_ISWLOWER,
BUILT_IN_ISWPRINT,
BUILT_IN_ISWPUNCT,
BUILT_IN_ISWSPACE,
BUILT_IN_ISWUPPER,
BUILT_IN_ISWXDIGIT,
BUILT_IN_TOWLOWER,
BUILT_IN_TOWUPPER,
BUILT_IN_ABORT,
BUILT_IN_ABS,
BUILT_IN_AGGREGATE_INCOMING_ADDRESS,
BUILT_IN_ALLOCA,
BUILT_IN_APPLY,
BUILT_IN_APPLY_ARGS,
BUILT_IN_ARGS_INFO,
BUILT_IN_CALLOC,
BUILT_IN_CLASSIFY_TYPE,
BUILT_IN_CLZ,
BUILT_IN_CLZIMAX,
BUILT_IN_CLZL,
BUILT_IN_CLZLL,
BUILT_IN_CONSTANT_P,
BUILT_IN_CTZ,
BUILT_IN_CTZIMAX,
BUILT_IN_CTZL,
BUILT_IN_CTZLL,
BUILT_IN_DCGETTEXT,
BUILT_IN_DGETTEXT,
BUILT_IN_DWARF_CFA,
BUILT_IN_DWARF_SP_COLUMN,
BUILT_IN_EH_RETURN,
BUILT_IN_EH_RETURN_DATA_REGNO,
BUILT_IN_EXECL,
BUILT_IN_EXECLP,
BUILT_IN_EXECLE,
BUILT_IN_EXECV,
BUILT_IN_EXECVP,
BUILT_IN_EXECVE,
BUILT_IN_EXIT,
BUILT_IN_EXPECT,
BUILT_IN_EXTEND_POINTER,
BUILT_IN_EXTRACT_RETURN_ADDR,
BUILT_IN_FFS,
BUILT_IN_FFSIMAX,
BUILT_IN_FFSL,
BUILT_IN_FFSLL,
BUILT_IN_FORK,
BUILT_IN_FRAME_ADDRESS,
BUILT_IN_FROB_RETURN_ADDR,
BUILT_IN_GETTEXT,
BUILT_IN_IMAXABS,
BUILT_IN_INIT_DWARF_REG_SIZES,
BUILT_IN_FINITE,
BUILT_IN_FINITEF,
BUILT_IN_FINITEL,
BUILT_IN_ISINF,
BUILT_IN_ISINFF,
BUILT_IN_ISINFL,
BUILT_IN_ISNAN,
BUILT_IN_ISNANF,
BUILT_IN_ISNANL,
BUILT_IN_ISGREATER,
BUILT_IN_ISGREATEREQUAL,
BUILT_IN_ISLESS,
BUILT_IN_ISLESSEQUAL,
BUILT_IN_ISLESSGREATER,
BUILT_IN_ISUNORDERED,
BUILT_IN_LABS,
BUILT_IN_LLABS,
BUILT_IN_LONGJMP,
BUILT_IN_MALLOC,
BUILT_IN_NEXT_ARG,
BUILT_IN_PARITY,
BUILT_IN_PARITYIMAX,
BUILT_IN_PARITYL,
BUILT_IN_PARITYLL,
BUILT_IN_POPCOUNT,
BUILT_IN_POPCOUNTIMAX,
BUILT_IN_POPCOUNTL,
BUILT_IN_POPCOUNTLL,
BUILT_IN_PREFETCH,
BUILT_IN_RETURN,
BUILT_IN_RETURN_ADDRESS,
BUILT_IN_SAVEREGS,
BUILT_IN_SETJMP,
BUILT_IN_STDARG_START,
BUILT_IN_STRFMON,
BUILT_IN_STRFTIME,
BUILT_IN_TRAP,
BUILT_IN_UNWIND_INIT,
BUILT_IN_UPDATE_SETJMP_BUF,
BUILT_IN_VA_COPY,
BUILT_IN_VA_END,
BUILT_IN_VA_START,
BUILT_IN__EXIT,
BUILT_IN__EXIT2,
BUILT_IN_INIT_TRAMPOLINE,
BUILT_IN_ADJUST_TRAMPOLINE,
BUILT_IN_NONLOCAL_GOTO,
BUILT_IN_STACK_SAVE,
BUILT_IN_STACK_RESTORE,
BUILT_IN_PROFILE_FUNC_ENTER,
BUILT_IN_PROFILE_FUNC_EXIT,
  BUILT_IN_COMPLEX_MUL_MIN,
  BUILT_IN_COMPLEX_MUL_MAX
    = BUILT_IN_COMPLEX_MUL_MIN
      + MAX_MODE_COMPLEX_FLOAT
      - MIN_MODE_COMPLEX_FLOAT,

  BUILT_IN_COMPLEX_DIV_MIN,
  BUILT_IN_COMPLEX_DIV_MAX
    = BUILT_IN_COMPLEX_DIV_MIN
      + MAX_MODE_COMPLEX_FLOAT
      - MIN_MODE_COMPLEX_FLOAT,

  END_BUILTINS
};


struct function;
struct tree_decl
{
  struct tree_common common;
  location_t locus;
  unsigned int uid;
  tree size;
  __extension__ enum machine_mode mode : 8;

  unsigned external_flag : 1;
  unsigned nonlocal_flag : 1;
  unsigned regdecl_flag : 1;
  unsigned inline_flag : 1;
  unsigned bit_field_flag : 1;
  unsigned virtual_flag : 1;
  unsigned ignored_flag : 1;
  unsigned abstract_flag : 1;

  unsigned in_system_header_flag : 1;
  unsigned common_flag : 1;
  unsigned defer_output : 1;
  unsigned transparent_union : 1;
  unsigned static_ctor_flag : 1;
  unsigned static_dtor_flag : 1;
  unsigned artificial_flag : 1;
  unsigned weak_flag : 1;

  unsigned non_addr_const_p : 1;
  unsigned no_instrument_function_entry_exit : 1;
  unsigned comdat_flag : 1;
  unsigned malloc_flag : 1;
  unsigned no_limit_stack : 1;
  __extension__ enum built_in_class built_in_class : 2;
  unsigned pure_flag : 1;

  unsigned non_addressable : 1;
  unsigned user_align : 1;
  unsigned uninlinable : 1;
  unsigned thread_local_flag : 1;
  unsigned declared_inline_flag : 1;
  unsigned seen_in_bind_expr : 1;
  __extension__ enum symbol_visibility visibility : 2;
  unsigned visibility_specified : 1;

  unsigned lang_flag_0 : 1;
  unsigned lang_flag_1 : 1;
  unsigned lang_flag_2 : 1;
  unsigned lang_flag_3 : 1;
  unsigned lang_flag_4 : 1;
  unsigned lang_flag_5 : 1;
  unsigned lang_flag_6 : 1;
  unsigned lang_flag_7 : 1;

  unsigned possibly_inlined : 1;
  unsigned preserve_flag: 1;
  unsigned gimple_formal_temp : 1;
  unsigned debug_expr_is_from : 1;


  union tree_decl_u1 {


    enum built_in_function f;


    long i;


    struct tree_decl_u1_a {
      unsigned int align : 24;
      unsigned int off_align : 8;
    } a;
  } u1;

  tree size_unit;
  tree name;
  tree context;
  tree arguments;
  tree result;
  tree initial;
  tree abstract_origin;
  tree assembler_name;
  tree section_name;
  tree attributes;
  rtx rtl;
  union tree_decl_u2 {
    struct function * f;
    rtx r;
    tree t;
    int i;
  } u2;

  tree saved_tree;
  tree vindex;
  long pointer_alias_set;
  struct lang_decl *lang_specific;
};

struct tree_statement_list_node

{
  struct tree_statement_list_node *prev;
  struct tree_statement_list_node *next;
  tree stmt;
};

struct tree_statement_list

{
  struct tree_common common;
  struct tree_statement_list_node *head;
  struct tree_statement_list_node *tail;
};

struct value_set;

struct tree_value_handle
{
  struct tree_common common;
  struct value_set * expr_set;
  unsigned int id;
};



struct tree_block
{
  struct tree_common common;
  unsigned handler_block_flag : 1;
  unsigned abstract_flag : 1;
  unsigned block_num : 30;
  tree vars;
  tree subblocks;
  tree supercontext;
  tree abstract_origin;
  tree fragment_origin;
  tree fragment_chain;
};

struct die_struct;

struct tree_type
{
  struct tree_common common;
  tree values;
  tree size;
  tree size_unit;
  tree attributes;
  unsigned int uid;
  unsigned int precision : 9;
  __extension__ enum machine_mode mode : 7;
  unsigned string_flag : 1;
  unsigned no_force_blk_flag : 1;
  unsigned needs_constructing_flag : 1;
  unsigned transparent_union_flag : 1;
  unsigned packed_flag : 1;
  unsigned restrict_flag : 1;
  unsigned contains_placeholder_bits : 2;
  unsigned lang_flag_0 : 1;
  unsigned lang_flag_1 : 1;
  unsigned lang_flag_2 : 1;
  unsigned lang_flag_3 : 1;
  unsigned lang_flag_4 : 1;
  unsigned lang_flag_5 : 1;
  unsigned lang_flag_6 : 1;
  unsigned user_align : 1;
  unsigned int align;
  tree pointer_to;
  tree reference_to;
  union tree_type_symtab {
    int address;
    char * pointer;
    struct die_struct * die;
  } symtab;
  tree name;
  tree minval;
  tree maxval;
  tree next_variant;
  tree main_variant;
  tree binfo;
  tree context;
  long alias_set;
  struct lang_type *lang_specific;
};

typedef struct VEC_tree { unsigned num; unsigned alloc; tree vec[1]; } VEC_tree; 
struct tree_binfo
{
  struct tree_common common;
  tree offset;
  tree vtable;
  tree virtuals;
  tree vptr_field;
  VEC_tree *base_accesses;
  tree inheritance;
  tree vtt_subvtt;
  tree vtt_vptr;
  VEC_tree base_binfos;
};



struct ptr_info_def;
struct edge_def;



struct tree_ssa_name
{
  struct tree_common common;
  tree var;
  unsigned int version;
  struct ptr_info_def *ptr_info;
  tree value_handle;
  void * aux;
};


struct phi_arg_d
{
  tree def;
  unsigned char nonzero;
};

struct tree_phi_node
{
  struct tree_common common;
  tree result;
  int num_args;
  int capacity;
  int rewritten;
  struct basic_block_def *bb;
  struct dataflow_d *df;
  struct phi_arg_d a[1];
};

union tree_node

{
  struct tree_common common;
  struct tree_int_cst int_cst;
  struct tree_real_cst real_cst;
  struct tree_vector vector;
  struct tree_string string;
  struct tree_complex complex;
  struct tree_identifier identifier;
  struct tree_decl decl;
  struct tree_type type;
  struct tree_list list;
  struct tree_vec vec;
  struct tree_exp exp;
  struct tree_ssa_name ssa_name;
  struct tree_phi_node phi;
  struct tree_block block;
  struct tree_binfo binfo;
  struct tree_statement_list stmt_list;
  struct tree_value_handle value_handle;
};



enum tree_dump_index
{
  TDI_none,
  TDI_tu,
  TDI_class,
  TDI_original,
  TDI_generic,
  TDI_nested,
  TDI_inlined,

  TDI_vcg,

  TDI_tree_all,
  TDI_rtl_all,
  TDI_ipa_all,

  TDI_cgraph,

  DFI_MIN,
  DFI_sibling = DFI_MIN,
  DFI_eh,
  DFI_jump,
  DFI_cse,
  DFI_gcse,
  DFI_loop,
  DFI_bypass,
  DFI_cfg,
  DFI_bp,
  DFI_vpt,
  DFI_ce1,
  DFI_tracer,
  DFI_loop2,
  DFI_web,
  DFI_cse2,
  DFI_life,
  DFI_combine,
  DFI_ce2,
  DFI_regmove,
  DFI_sms,
  DFI_sched,
  DFI_lreg,
  DFI_greg,
  DFI_postreload,
  DFI_gcse2,
  DFI_flow2,
  DFI_peephole2,
  DFI_ce3,
  DFI_rnreg,
  DFI_bbro,
  DFI_branch_target_load,
  DFI_sched2,
  DFI_stack,
  DFI_vartrack,
  DFI_mach,
  DFI_dbr,

  TDI_end
};


extern const char *dump_base_name;

struct dump_file_info
{
  const char *suffix;
  const char *swtch;
  const char *glob;
  int flags;
  int state;
  int num;
  int letter;
};


//extern struct dump_file_info dump_files[TDI_end];


/// GCC STUFF
extern const char *const tree_code_name[];
#define TREE_CODE(NODE) ((enum tree_code) (NODE)->common.code)

/* This is the name of the object as written by the user.
   It is an IDENTIFIER_NODE.  */
//#define DECL_NAME(NODE) (DECL_CHECK (NODE)->decl.name)

tree DECL_NAME(tree NODE);

const char * IDENTIFIER_POINTER(tree NODE);

#define ATTRIBUTE_NORETURN

extern void tree_check_failed (const tree, const char *, int, const char *,
			       ...) ;
extern void tree_not_check_failed (const tree, const char *, int, const char *,
				   ...);

enum tree_code_class {
  tcc_exceptional,
  tcc_constant,
  tcc_type,
  tcc_declaration,
  tcc_reference,
  tcc_comparison,
  tcc_unary,
  tcc_binary,
  tcc_statement,
  tcc_expression
};

extern void tree_class_check_failed (const tree, const enum tree_code_class ,
				     const char *, int, const char *);

extern const enum tree_code_class tree_code_type[];

enum tree_code_class  TREE_CODE_CLASS(enum tree_code CODE);
tree  TREE_CLASS_CHECK(tree T, const enum tree_code_class CLASS);

tree DECL_CHECK(tree T);//		TREE_CLASS_CHECK (T, tcc_declaration)
//#define DECL_CHECK(T)		TREE_CLASS_CHECK (T, tcc_declaration)

tree  IDENTIFIER_NODE_CHECK(tree t);  //       TREE_CHECK (t, IDENTIFIER_NODE)
tree TREE_CHECK(tree T,  enum tree_code CODE);

/* Flags used with queue functions.  */
#define DUMP_NONE     0
#define DUMP_BINFO    1

/*
extern splay_tree splay_tree_new (splay_tree_compare_fn, 
				  splay_tree_delete_key_fn, 
				  splay_tree_delete_value_fn);
*/
extern int splay_tree_compare_pointers (splay_tree_key, splay_tree_key);
typedef void (*splay_tree_delete_value_fn) (splay_tree_value);




// from tree.h
#define TDF_ADDRESS	(1 << 0)	/* dump node addresses */
#define TDF_SLIM	(1 << 1)	/* don't go wild following links */
#define TDF_RAW  	(1 << 2)	/* don't unparse the function */
#define TDF_DETAILS	(1 << 3)	/* show more detailed info about
					   each pass */
#define TDF_STATS	(1 << 4)	/* dump various statistics about
					   each pass */
#define TDF_BLOCKS	(1 << 5)	/* display basic block boundaries */
#define TDF_VOPS	(1 << 6)	/* display virtual operands */
#define TDF_LINENO	(1 << 7)	/* display statement line numbers */
#define TDF_UID		(1 << 8)	/* display decl UIDs */

#define TDF_TREE	(1 << 9)	/* is a tree dump */
#define TDF_RTL		(1 << 10)	/* is a RTL dump */
#define TDF_IPA		(1 << 11)	/* is an IPA dump */


// data


#define EDGE_FALLTHRU		1	/* 'Straight line' flow */
#define EDGE_ABNORMAL		2	/* Strange flow, like computed
					   label, or eh */
#define EDGE_ABNORMAL_CALL	4	/* Call with abnormal exit
					   like an exception, or sibcall */
#define EDGE_EH			8	/* Exception throw */
#define EDGE_FAKE		16	/* Not a real edge (profile.c) */
#define EDGE_DFS_BACK		32	/* A backwards edge */
#define EDGE_CAN_FALLTHRU	64	/* Candidate for straight line
					   flow.  */
#define EDGE_IRREDUCIBLE_LOOP	128	/* Part of irreducible loop.  */
#define EDGE_SIBCALL		256	/* Edge from sibcall to exit.  */
#define EDGE_LOOP_EXIT		512	/* Exit of a loop.  */
#define EDGE_TRUE_VALUE		1024	/* Edge taken when controlling
					   predicate is nonzero.  */
#define EDGE_FALSE_VALUE	2048	/* Edge taken when controlling
					   predicate is zero.  */
#define EDGE_EXECUTABLE		4096	/* Edge is executable.  Only
					   valid during SSA-CCP.  */
#define EDGE_CROSSING		8192    /* Edge crosses between hot
					   and cold sections, when we
					   do partitioning.  */
#define EDGE_ALL_FLAGS	       16383
#define EDGE_COMPLEX	(EDGE_ABNORMAL | EDGE_ABNORMAL_CALL | EDGE_EH)



#define FOR_EACH_EDGE(EDGE,ITER,EDGE_VEC) \
  for ((EDGE) = NULL, (ITER) = ei_start ((EDGE_VEC)); \
       ((EDGE) = ei_safe_edge ((ITER))); \
       ei_next (&(ITER)))

/* For iterating over basic blocks.  */
#define FOR_BB_BETWEEN(BB, FROM, TO, DIR) \
  for (BB = FROM; BB != TO; BB = BB->DIR)

#define FOR_EACH_BB(BB) \
  FOR_BB_BETWEEN (BB, ENTRY_BLOCK_PTR->next_bb, EXIT_BLOCK_PTR, next_bb)



typedef struct basic_block_def *basic_block;
extern basic_block ENTRY_BLOCK_PTR;
extern basic_block EXIT_BLOCK_PTR;
extern void check_bb_profile (basic_block, FILE *);


////////////////////////////////////////////////
// from introspector_gcc_data.c

typedef enum
{
DK_FATAL,
DK_ICE,
DK_ERROR,
DK_SORRY,
DK_WARNING,
DK_ANACHRONISM,
DK_NOTE,
DK_DEBUG,
  DK_LAST_DIAGNOSTIC_KIND
} diagnostic_t;

typedef struct
{
  const char *format_spec;
  va_list *args_ptr;
  int err_no;
} text_info;

typedef struct
{
  text_info message;
  location_t location;
  diagnostic_t kind;
} diagnostic_info;

typedef struct diagnostic_context diagnostic_context;
typedef void (*diagnostic_starter_fn) (diagnostic_context *,
                                       diagnostic_info *);
typedef diagnostic_starter_fn diagnostic_finalizer_fn;



struct obstack;
struct _obstack_chunk
{
  char *limit;
  struct _obstack_chunk *prev;
  char contents[4];
};

struct obstack
{
  long chunk_size;
  struct _obstack_chunk *chunk;
  char *object_base;
  char *next_free;
  char *chunk_limit;
  int temp;
  int alignment_mask;
  struct _obstack_chunk *(*chunkfun) (void *, long);
  void (*freefun) (void *, struct _obstack_chunk *);
  void *extra_arg;
  unsigned use_extra_arg:1;
  unsigned maybe_empty_object:1;
  unsigned alloc_failed:1;
};

typedef struct
{
  struct obstack obstack;
  FILE *stream;
  int line_length;
  char digit_buffer[128];
} output_buffer;

typedef unsigned int pp_flags;


typedef enum
{
  pp_none, pp_before, pp_after
} pp_padding;


typedef enum
{
  DIAGNOSTICS_SHOW_PREFIX_ONCE = 0x0,
  DIAGNOSTICS_SHOW_PREFIX_NEVER = 0x1,
  DIAGNOSTICS_SHOW_PREFIX_EVERY_LINE = 0x2
} diagnostic_prefixing_rule_t;

typedef struct pretty_print_info pretty_printer;
typedef unsigned char (*printer_fn) (pretty_printer *, text_info *);

struct pretty_print_info
{
  output_buffer *buffer;
  const char *prefix;
  pp_padding padding;
  int maximum_length;
  int ideal_maximum_length;
  int indent_skip;
  diagnostic_prefixing_rule_t prefixing_rule;
  printer_fn format_decoder;
  unsigned char emitted_prefix;
  unsigned char need_newline;
};





struct diagnostic_context
{
  pretty_printer *printer;
  int diagnostic_count[DK_LAST_DIAGNOSTIC_KIND];
  unsigned char issue_warnings_are_errors_message;
  unsigned char warning_as_error_requested;
  unsigned char abort_on_error;
  diagnostic_starter_fn begin_diagnostic;
  diagnostic_finalizer_fn end_diagnostic;
  void (*internal_error) (const char *, va_list *);
  tree last_function;
  int last_module;
  int lock;
};

typedef void (*lang_print_tree_hook) (FILE *, tree, int indent);

struct pointer_set_t;

struct lang_hooks_for_tree_inlining
{
  tree (*walk_subtrees) (tree *, int *,
                         tree (*) (tree *, int *, void *),
                         void *, struct pointer_set_t*);
  int (*cannot_inline_tree_fn) (tree *);
  int (*disregard_inline_limits) (tree);
  tree (*add_pending_fn_decls) (void *, tree);
  int (*auto_var_in_fn_p) (tree, tree);
  int (*anon_aggr_type_p) (tree);
  unsigned char (*var_mod_type_p) (tree, tree);
  int (*start_inlining) (tree);
  void (*end_inlining) (tree);
  tree (*convert_parm_for_inlining) (tree, tree, tree, int);
};

struct lang_hooks_for_callgraph
{
  tree (*analyze_expr) (tree *, int *, tree);
  void (*expand_function) (tree);
};

struct lang_hooks_for_functions
{  void (*init) (struct function *);
  void (*final) (struct function *);
  void (*enter_nested) (struct function *);
  void (*leave_nested) (struct function *);
  unsigned char (*missing_noreturn_ok_p) (tree);
};

struct lang_hooks_for_tree_dump
{
  unsigned char (*dump_tree) (void *, tree);
  int (*type_quals) (tree);
};

struct lang_hooks_for_types
{
  tree (*make_type) (enum tree_code);  tree (*type_for_mode) (enum machine_mode, int);
  tree (*type_for_size) (unsigned, int);
  tree (*unsigned_type) (tree);  
    tree (*signed_type) (tree);
  tree (*signed_or_unsigned_type) (int, tree);
  tree (*type_promotes_to) (tree);
  void (*register_builtin_type) (tree, const char *);
  void (*incomplete_type_error) (tree value, tree type);
  tree (*max_size) (tree);
  unsigned char hash_types;
};

struct lang_hooks_for_decls
{
  int (*global_bindings_p) (void);
  void (*insert_block) (tree);
  tree (*pushdecl) (tree);
  tree (*getdecls) (void);
  unsigned char (*warn_unused_global) (tree);
  void (*final_write_globals) (void);
  void (*prepare_assemble_variable) (tree);
  unsigned char (*ok_for_sibcall) (tree);
  const char * (*comdat_group) (tree);
};

struct lang_hooks
{
  const char *name;
  size_t identifier_size;
  size_t (*tree_size) (enum tree_code);
  unsigned int (*init_options) (unsigned int argc, const char **argv);
  void (*initialize_diagnostics) (struct diagnostic_context *);
  int (*handle_option) (size_t code, const char *arg, int value);
  unsigned char (*missing_argument) (const char *opt, size_t code);
  unsigned char (*post_options) (const char **);
  unsigned char (*init) (void);
  void (*finish) (void);  void (*parse_file) (int);
  void (*clear_binding_stack) (void);
  long (*get_alias_set) (tree);
  tree (*expand_constant) (tree);
  rtx (*expand_expr) (tree, rtx, enum machine_mode, int, rtx *);
  int (*expand_decl) (tree);
  tree (*truthvalue_conversion) (tree);
  int (*safe_from_p) (rtx, tree);
  void (*finish_incomplete_decl) (tree);
  unsigned char (*mark_addressable) (tree);
  tree (*staticp) (tree);
  void (*dup_lang_specific_decl) (tree);
  void (*set_decl_assembler_name) (tree);
  unsigned char (*can_use_bit_fields_p) (void);
  unsigned char reduce_bit_field_operations;
  unsigned char no_body_blocks;
  void (*print_statistics) (void);
  lang_print_tree_hook print_xnode;
  lang_print_tree_hook print_decl;
  lang_print_tree_hook print_type;
  lang_print_tree_hook print_identifier;
  const char *(*decl_printable_name) (tree decl, int verbosity);
  int (*types_compatible_p) (tree x, tree y);
  tree (*lang_get_callee_fndecl) (tree);
  void (*print_error_function) (struct diagnostic_context *, const char *);
  tree (*expr_size) (tree);
  long (*to_target_charset) (long);
  const struct attribute_spec *attribute_table;
  const struct attribute_spec *common_attribute_table;
  const struct attribute_spec *format_attribute_table;
  struct lang_hooks_for_functions function;
  struct lang_hooks_for_tree_inlining tree_inlining;
  struct lang_hooks_for_callgraph callgraph;
  struct lang_hooks_for_tree_dump tree_dump;
  struct lang_hooks_for_decls decls;
  struct lang_hooks_for_types types;
  int (*gimplify_expr) (tree *, tree *, tree *);
  tree (*fold_obj_type_ref) (tree, tree);
  tree (*builtin_function) (const char *name, tree type, int function_code,
                            enum built_in_class bt_class,
                            const char *library_name, tree attrs);
};


extern const struct lang_hooks lang_hooks;

///////////////////////////////////////////////// BASIC BLOCk

typedef struct edge_def *edge;
typedef struct VEC_edge { unsigned num; unsigned alloc; edge vec[1]; } VEC_edge; 


typedef struct {
  unsigned index;
  VEC_edge **container;
} edge_iterator;

typedef long long gcov_type;

struct loop;



typedef struct bitmap_obstack
{
  struct bitmap_element_def *elements;
  struct bitmap_head_def *heads;
  struct obstack obstack;
} bitmap_obstack;


typedef unsigned long BITMAP_WORD;

typedef struct bitmap_element_def
{
  struct bitmap_element_def *next;
  struct bitmap_element_def *prev;
  unsigned int indx;
  BITMAP_WORD bits[((128 + (8 * 4 * 1u) - 1) / (8 * 4 * 1u))];
} bitmap_element;

typedef struct bitmap_head_def {
  bitmap_element *first;
  bitmap_element *current;
  unsigned int indx;
  bitmap_obstack *obstack;
} bitmap_head;

typedef struct bitmap_head_def *bitmap;

/// basic block
struct basic_block_def
{
  rtx head_;
  rtx end_;
  tree stmt_list;
  VEC_edge *preds;
  VEC_edge *succs;
  bitmap global_live_at_start;
  bitmap global_live_at_end;
  void * aux;
  struct loop * loop_father;
  struct et_node * dom[2];
  struct basic_block_def *prev_bb;
  struct basic_block_def *next_bb;
  struct reorder_block_def * rbi;
  struct bb_ann_d *tree_annotations;
  gcov_type count;
  int index;
  int loop_depth;
  int frequency;
  int flags;
};

struct edge_def
{
  struct basic_block_def *src;
  struct basic_block_def *dest;
  union edge_def_insns {
    rtx r;
    tree t;
  } insns;
  void * aux;
  source_locus goto_locus;
  int flags;
  int probability;
  gcov_type count;
  unsigned int dest_idx;
};








extern tree current_function_decl;
extern int n_basic_blocks;
extern int last_basic_block;
extern int n_edges;


/* Structure to hold decision about unrolling/peeling.  */
enum lpt_dec
{
  LPT_NONE,
  LPT_PEEL_COMPLETELY,
  LPT_PEEL_SIMPLE,
  LPT_UNROLL_CONSTANT,
  LPT_UNROLL_RUNTIME,
  LPT_UNROLL_STUPID
};

struct lpt_decision
{
  enum lpt_dec decision;
  unsigned times;
};

typedef struct simple_bitmap_def
{
  unsigned int n_bits;
  unsigned int size;
  unsigned int bytes;
  unsigned long elms[1];
} *sbitmap;

// TODO 
#define bool unsigned char 
#define true 1
#define false 0

struct loop
{
  /* Index into loops array.  */
  int num;

  /* Basic block of loop header.  */
  basic_block header;

  /* Basic block of loop latch.  */
  basic_block latch;

  /* Basic block of loop preheader or NULL if it does not exist.  */
  basic_block pre_header;

  /* For loop unrolling/peeling decision.  */
  struct lpt_decision lpt_decision;

  /* Number of loop insns.  */
  unsigned ninsns;

  /* Average number of executed insns per iteration.  */
  unsigned av_ninsns;

  /* Array of edges along the preheader extended basic block trace.
     The source of the first edge is the root node of preheader
     extended basic block, if it exists.  */
  edge *pre_header_edges;

  /* Number of edges along the pre_header extended basic block trace.  */
  int num_pre_header_edges;

  /* The first block in the loop.  This is not necessarily the same as
     the loop header.  */
  basic_block first;

  /* The last block in the loop.  This is not necessarily the same as
     the loop latch.  */
  basic_block last;

  /* Number of blocks contained within the loop.  */
  unsigned num_nodes;

  /* Array of edges that enter the loop.  */
  edge *entry_edges;

  /* Number of edges that enter the loop.  */
  int num_entries;

  /* Array of edges that exit the loop.  */
  edge *exit_edges;

  /* Number of edges that exit the loop.  */
  int num_exits;

  /* Bitmap of blocks that dominate all exits of the loop.  */
  sbitmap exits_doms;

  /* The loop nesting depth.  */
  int depth;

  /* Superloops of the loop.  */
  struct loop **pred;

  /* The height of the loop (enclosed loop levels) within the loop
     hierarchy tree.  */
  int level;

  /* The outer (parent) loop or NULL if outermost loop.  */
  struct loop *outer;

  /* The first inner (child) loop or NULL if innermost loop.  */
  struct loop *inner;

  /* Link to the next (sibling) loop.  */
  struct loop *next;

  /* Loop that is copy of this loop.  */
  struct loop *copy;

  /* Nonzero if the loop is invalid (e.g., contains setjmp.).  */
  int invalid;

  /* Auxiliary info specific to a pass.  */
  void *aux;

  /* The following are currently used by loop.c but they are likely to
     disappear when loop.c is replaced and removed.  */

  /* The NOTE_INSN_LOOP_BEG.  */
  rtx start;

  /* The NOTE_INSN_LOOP_END.  */
  rtx end;

  /* For a rotated loop that is entered near the bottom,
     this is the label at the top.  Otherwise it is zero.  */
  rtx top;

  /* Place in the loop where control enters.  */
  rtx scan_start;

  /* The position where to sink insns out of the loop.  */
  rtx sink;

  /* List of all LABEL_REFs which refer to code labels outside the
     loop.  Used by routines that need to know all loop exits, such as
     final_biv_value and final_giv_value.

     This does not include loop exits due to return instructions.
     This is because all bivs and givs are pseudos, and hence must be
     dead after a return, so the presence of a return does not affect
     any of the optimizations that use this info.  It is simpler to
     just not include return instructions on this list.  */
  rtx exit_labels;

  /* The number of LABEL_REFs on exit_labels for this loop and all
     loops nested inside it.  */
  int exit_count;

  /* The probable number of times the loop is executed at runtime.
     This is an INTEGER_CST or an expression containing symbolic
     names.  Don't access this field directly:
     number_of_iterations_in_loop computes and caches the computed
     information in this field.  */
  tree nb_iterations;

  /* An INTEGER_CST estimation of the number of iterations.  NULL_TREE
     if there is no estimation.  */
  tree estimated_nb_iterations;

  /* Upper bound on number of iterations of a loop.  */
  struct nb_iter_bound *bounds;

  /* If not NULL, loop has just single exit edge stored here (edges to the
     EXIT_BLOCK_PTR do not count.  */
  edge single_exit;

  /* True when the loop does not carry data dependences, and
     consequently the iterations can be executed in any order.  False
     when the loop carries data dependences, or when the property is
     not decidable.  */
  bool parallel_p;
};




struct ptr_info_def
{


  unsigned int pt_anything : 1;


  unsigned int pt_malloc : 1;


  unsigned int value_escapes_p : 1;


  unsigned int is_dereferenced : 1;


  unsigned int pt_global_mem : 1;


  unsigned int pt_null : 1;


  bitmap pt_vars;





  tree name_mem_tag;
};





enum tree_ann_type { TREE_ANN_COMMON, VAR_ANN, STMT_ANN };

struct tree_ann_common_d
{
  enum tree_ann_type type;
  void * aux;
  tree value_handle;
};



enum need_phi_state {
  NEED_PHI_STATE_UNKNOWN,
  NEED_PHI_STATE_NO,
  NEED_PHI_STATE_MAYBE
};

enum mem_tag_kind {
  NOT_A_TAG,
  TYPE_TAG,
  NAME_TAG
};
struct rtvec_def;
typedef struct rtvec_def *rtvec;

typedef union varray_data_tag {
  char c[1];
  unsigned char uc[1];
  short s[1];
  unsigned short us[1];
  int i[1];
  unsigned int u[1];
  long l[1];
  unsigned long ul[1];
  long hint[1];
  unsigned long uhint[1];
  void * generic[1];
  void * generic_nogc[1];
  char * cptr[1];
  rtx rtx[1];
  rtvec rtvec[1];
  tree tree[1];
  struct bitmap_head_def * bitmap[1];
  struct reg_info_def * reg[1];
  struct basic_block_def * bb[1];
  struct elt_list * te[1];
  struct edge_def * e[1];
  tree * tp[1];
} varray_data;

enum varray_data_enum {
  VARRAY_DATA_C,
  VARRAY_DATA_UC,
  VARRAY_DATA_S,
  VARRAY_DATA_US,
  VARRAY_DATA_I,
  VARRAY_DATA_U,
  VARRAY_DATA_L,
  VARRAY_DATA_UL,
  VARRAY_DATA_HINT,
  VARRAY_DATA_UHINT,
  VARRAY_DATA_GENERIC,
  VARRAY_DATA_GENERIC_NOGC,
  VARRAY_DATA_CPTR,
  VARRAY_DATA_RTX,
  VARRAY_DATA_RTVEC,
  VARRAY_DATA_TREE,
  VARRAY_DATA_BITMAP,
  VARRAY_DATA_REG,
  VARRAY_DATA_BB,
  VARRAY_DATA_TE,
  VARRAY_DATA_EDGE,
  VARRAY_DATA_TREE_PTR,
  NUM_VARRAY_DATA
};

struct varray_head_tag {
  size_t num_elements;
  size_t elements_used;
  enum varray_data_enum type;
  const char *name;
  varray_data data;
};

typedef struct varray_head_tag *varray_type;

struct dataflow_d
{
  varray_type immediate_uses;
  tree uses[2];
};

typedef struct dataflow_d *dataflow_t;

typedef struct stmt_operands_d
{
  struct def_optype_d * def_ops;
  struct use_optype_d * use_ops;
  struct v_may_def_optype_d * v_may_def_ops;
  struct vuse_optype_d * vuse_ops;
  struct v_must_def_optype_d * v_must_def_ops;
} stmt_operands_t;

struct stmt_ann_d
{
  struct tree_ann_common_d common;
  unsigned modified : 1;
  unsigned makes_aliased_loads : 1;
  unsigned makes_aliased_stores : 1;
  unsigned has_volatile_ops : 1;
  unsigned makes_clobbering_call : 1;
  basic_block bb;
  struct stmt_operands_d operands;
  dataflow_t df;
  bitmap addresses_taken;
  unsigned int uid;
};


struct var_ann_d
{
  struct tree_ann_common_d common;
  unsigned out_of_ssa_tag : 1;
  unsigned root_var_processed : 1;
  __extension__ enum mem_tag_kind mem_tag_kind : 2;
  unsigned is_alias_tag : 1;
  unsigned used : 1;
  __extension__ enum need_phi_state need_phi_state : 2;
  unsigned in_vuse_list : 1;
  unsigned in_v_may_def_list : 1;
  tree type_mem_tag;
  varray_type may_aliases;
  size_t uid;
  unsigned partition;
  unsigned root_index;
  tree default_def;
  tree current_def;
};

union tree_ann_d
{
  struct tree_ann_common_d common;
  struct var_ann_d decl;
  struct stmt_ann_d stmt;
};


extern tree first_stmt (basic_block);
extern tree last_stmt (basic_block);
extern varray_type basic_block_info;
extern struct function *cfun;


struct bb_ann_d
{
  tree phi_nodes;
  unsigned has_escape_site : 1;
  unsigned incoming_edge_threaded : 1;
  struct edge_prediction *predictions;
};

typedef struct bb_ann_d *bb_ann_t;

typedef struct ix86_args {
  int words;
  int nregs;
  int regno;
  int fastcall;
  int sse_words;
  int sse_nregs;
  int warn_sse;
  int warn_mmx;
  int sse_regno;
  int mmx_words;
  int mmx_nregs;
  int mmx_regno;
  int maybe_vaarg;
} CUMULATIVE_ARGS;

struct function
{
  struct eh_status *eh;
  struct expr_status *expr;
  struct emit_status *emit;
  struct varasm_status *varasm;
  tree saved_tree;
  tree saved_args;
  tree saved_static_chain_decl;
  tree decl;
  struct function *outer;
  int pops_args;
  int args_size;
  int pretend_args_size;
  int outgoing_args_size;
  rtx arg_offset_rtx;
  CUMULATIVE_ARGS args_info;
  rtx return_rtx;
  rtx internal_arg_pointer;
  struct initial_value_struct *hard_reg_initial_vals;
  rtx x_nonlocal_goto_handler_labels;
  rtx x_return_label;
  rtx x_naked_return_label;
  rtx x_stack_slot_list;
  rtx x_tail_recursion_reentry;
  rtx x_arg_pointer_save_area;
  long x_frame_offset;
  tree static_chain_decl;
  tree nonlocal_goto_save_area;
  rtx x_parm_birth_insn;
  struct varray_head_tag * x_used_temp_slots;
  struct temp_slot *x_avail_temp_slots;
  int x_temp_slot_level;
  struct var_refs_queue *fixup_var_refs_queue;
  int inlinable;
  int no_debugging_symbols;
  rtvec original_arg_vector;
  tree original_decl_initial;
  int inl_max_label_num;
  int funcdef_no;
  int max_loop_depth;
  struct machine_function * machine;
  unsigned int stack_alignment_needed;
  unsigned int preferred_stack_boundary;
  unsigned char recursive_call_emit;
  unsigned char tail_call_emit;
  struct language_function * language;
  rtx epilogue_delay_list;
  enum function_frequency {
    FUNCTION_FREQUENCY_UNLIKELY_EXECUTED,
    FUNCTION_FREQUENCY_NORMAL,
    FUNCTION_FREQUENCY_HOT
  } function_frequency;
  int max_jumptable_ents;
  int last_label_uid;
  location_t function_end_locus;
  struct varray_head_tag *ib_boundaries_block;
  tree unexpanded_var_list;
  unsigned int returns_struct : 1;
  unsigned int returns_pcc_struct : 1;
  unsigned int returns_pointer : 1;
  unsigned int calls_setjmp : 1;
  unsigned int calls_alloca : 1;
  unsigned int calls_eh_return : 1;
  unsigned int has_nonlocal_label : 1;
  unsigned int has_nonlocal_goto : 1;
  unsigned int contains_functions : 1;
  unsigned int is_thunk : 1;
  unsigned int all_throwers_are_sibcalls : 1;
  unsigned int instrument_entry_exit : 1;
  unsigned int profile : 1;
  unsigned int limit_stack : 1;
  unsigned int stdarg : 1;
  unsigned int x_dont_save_pending_sizes_p : 1;
  unsigned int uses_const_pool : 1;
  unsigned int uses_pic_offset_table : 1;
  unsigned int uses_eh_lsda : 1;
  unsigned int arg_pointer_save_area_init : 1;
};

extern struct function *cfun;
int get_lineno (tree);


void ei_next (edge_iterator *i);
edge_iterator ei_start (VEC_edge **ev);
edge ei_safe_edge (edge_iterator i);


extern void print_generic_stmt (FILE *, tree, int);
extern void print_generic_stmt_indented (FILE *, tree, int, int);
extern void print_generic_expr (FILE *, tree, int);
extern void print_generic_decl (FILE *, tree, int);

#define BASIC_BLOCK(N)  (VARRAY_BB (basic_block_info, (N)))


extern void varray_check_failed (varray_type, size_t, const char *, int,
				 const char *) ATTRIBUTE_NORETURN;

#define VARRAY_CHECK(VA, N, T) __extension__			\
(*({ varray_type const _va = (VA);				\
     const size_t _n = (N);					\
     if (_n >= _va->num_elements)				\
       varray_check_failed (_va, _n, __FILE__, __LINE__, __FUNCTION__);	\
     &_va->data.T[_n]; }))

#define VARRAY_BB(VA, N)		VARRAY_CHECK (VA, N, bb)

#define SCALE(x) ((unsigned long) ((x) < 1024*10	\
		  ? (x)					\
		  : ((x) < 1024*1024*10			\
		     ? (x) / 1024			\
		     : (x) / (1024*1024))))

#define LABEL(x) ((x) < 1024*10 ? 'b' : ((x) < 1024*1024*10 ? 'k' : 'M'))

#define ei_start(iter) ei_start_1 (&(iter))

//#define VEC_OP(TDEF,OP) VEC_OP_(VEC(TDEF),OP)
//#define VEC_OP_(VEC,OP) VEC_OP__(VEC,OP)
//#define VEC_OP__(VEC,OP) VEC ## _ ## OP


//#define VEC_length(TDEF,V)	(VEC_OP(TDEF,length)(V))
//#define EDGE_COUNT(ev)			VEC_length (edge, (ev))

extern const unsigned char tree_code_length[];
extern const char *const tree_code_class_strings[];

/// MACROS -> FUNCTIONS

char TREE_VEC_LENGTH(tree t); 
const char * TREE_CODE_CLASS_STRING(tree t); 
const char * TREE_STRING_POINTER(tree t); 
enum tree_code TREE_SET_CODE(tree t,enum tree_code a);
enum tree_code_class TREE_READONLY_DECL_P(tree t); 
int TREE_ADDRESSABLE(tree t); 
int TREE_CODE_LENGTH(tree t); 
int TREE_COMPLEXITY(tree t); 
int TREE_CONSTANT_OVERFLOW(tree t); 
int TREE_DEPRECATED(tree t); 
int TREE_HASH(tree t); 
int TREE_STRING_LENGTH(tree t); 
long TREE_INT_CST_HIGH(tree t); 
struct real_value * TREE_REAL_CST(tree t); 
struct real_value * TREE_REAL_CST_PTR(tree t); 
struct tree_int_cst_lowh * TREE_INT_CST(tree t); 
tree DECL_SAVED_TREE(tree t); 
tree PHI_ARG_DEF_TREE(tree t,int a);
tree PHI_RESULT_TREE(tree t); 
tree TREE_BINFO_CHECK(tree t); 
tree TREE_BLOCK(tree t); 
tree TREE_CHAIN(tree t); 
tree TREE_CONSTANT(tree t); 
tree TREE_IMAGPART(tree t); 
tree TREE_OPERAND(tree t, int a);
tree TREE_OPERAND_CHECK(tree t, int a);
tree TREE_OPERAND_CHECK_CODE(tree t,int b);
tree TREE_PURPOSE(tree t); 
tree TREE_REALPART(tree t); 
tree TREE_RTL_OPERAND_CHECK(tree t,int b);
tree TREE_TYPE(tree t); 
tree TREE_VALUE(tree t); 
tree TREE_VECTOR_CST_ELTS(tree t); 
tree TREE_VEC_ELT(tree t, int a);
tree TREE_VEC_ELT_CHECK(tree t,int a);
tree TREE_VEC_END(tree t); 
tree VARRAY_PUSH_TREE_PTR(varray_type t,int a);
tree VARRAY_TOP_TREE(varray_type t); 
tree VARRAY_TOP_TREE_PTR(varray_type t); 
tree VARRAY_TREE(varray_type t,int a);
tree VARRAY_TREE_INIT(varray_type t,int a, int b);
tree VARRAY_TREE_PTR(varray_type t,int a);
tree VARRAY_TREE_PTR_INIT(varray_type t,int a, int b);
unsigned TREE_ASM_WRITTEN(tree t); 
unsigned TREE_INVARIANT(tree t); 
unsigned TREE_LANG_FLAG_0(tree t); 
unsigned TREE_LANG_FLAG_1(tree t); 
unsigned TREE_LANG_FLAG_2(tree t); 
unsigned TREE_LANG_FLAG_3(tree t); 
unsigned TREE_LANG_FLAG_4(tree t); 
unsigned TREE_LANG_FLAG_5(tree t); 
unsigned TREE_LANG_FLAG_6(tree t); 
unsigned TREE_NOTHROW(tree t); 
unsigned TREE_NO_WARNING(tree t); 
unsigned TREE_OVERFLOW(tree t); 
unsigned TREE_PRIVATE(tree t); 
unsigned TREE_PROTECTED(tree t); 
unsigned TREE_PUBLIC(tree t); 
unsigned TREE_READONLY(tree t); 
unsigned TREE_SIDE_EFFECTS(tree t); 
unsigned TREE_STATIC(tree t); 
unsigned TREE_SYMBOL_REFERENCED(tree t); 
unsigned TREE_THIS_NOTRAP(tree t); 
unsigned TREE_THIS_VOLATILE(tree t); 
unsigned TREE_USED(tree t); 
unsigned TREE_VISITED(tree t); 
unsigned long TREE_INT_CST_LOW(tree t); 
void VARRAY_PUSH_TREE(varray_type t,int a);

// added this in
FILE *dump_begin (enum tree_dump_index, int *);
void dump_end (enum tree_dump_index, FILE *);
void dump_node (tree, int, FILE *);
void dump_function (enum tree_dump_index, tree);


typedef enum
{
TV_TOTAL,
TV_GC,


TV_DUMP,

TV_CGRAPH,
TV_CGRAPHOPT,

TV_CFG,

TV_CLEANUP_CFG,
TV_CFG_VERIFY,
TV_DELETE_TRIVIALLY_DEAD,

TV_LIFE,
TV_LIFE_UPDATE,

TV_ALIAS_ANALYSIS,
TV_REG_SCAN,
TV_REBUILD_JUMP,

TV_CPP,
TV_LEX,
TV_PARSE,
TV_NAME_LOOKUP,
TV_INTEGRATION,
TV_TREE_GIMPLIFY,
TV_TREE_EH,
TV_TREE_CFG,
TV_TREE_CLEANUP_CFG,
TV_TREE_VRP,
TV_TREE_COPY_PROP,
TV_TREE_STORE_COPY_PROP,
TV_FIND_REFERENCED_VARS,
TV_TREE_PTA,
TV_TREE_MAY_ALIAS,
TV_TREE_INSERT_PHI_NODES,
TV_TREE_SSA_REWRITE_BLOCKS,
TV_TREE_SSA_OTHER,
TV_TREE_SSA_INCREMENTAL,
TV_TREE_OPS,
TV_TREE_SSA_DOMINATOR_OPTS,
TV_TREE_SRA,
TV_TREE_STORE_CCP,
TV_TREE_CCP,
TV_TREE_SPLIT_EDGES,
TV_TREE_PRE,
TV_TREE_REDPHI,
TV_TREE_FRE,
TV_TREE_SINK,
TV_TREE_PHIOPT,
TV_TREE_FORWPROP,
TV_TREE_DCE,
TV_TREE_CD_DCE,
TV_TREE_DSE,
TV_TREE_MERGE_PHI,
TV_TREE_LOOP,
TV_TREE_LOOP_BOUNDS,
TV_LIM,
TV_TREE_LOOP_IVCANON,
TV_TREE_LOOP_UNSWITCH,
TV_COMPLETE_UNROLL,
TV_TREE_VECTORIZATION,
TV_TREE_LINEAR_TRANSFORM,
TV_TREE_LOOP_IVOPTS,
TV_TREE_LOOP_INIT,
TV_TREE_LOOP_FINI,
TV_TREE_CH,
TV_TREE_SSA_UNCPROP,
TV_TREE_SSA_TO_NORMAL,
TV_TREE_NRV,
TV_TREE_COPY_RENAME,
TV_TREE_SSA_VERIFY,
TV_TREE_STMT_VERIFY,
TV_CGRAPH_VERIFY,
TV_DOM_FRONTIERS,
TV_CONTROL_DEPENDENCES,
TV_OVERLOAD,
TV_TEMPLATE_INSTANTIATION,
TV_EXPAND,
TV_VARCONST,
TV_JUMP,
TV_CSE,
TV_LOOP,
TV_GCSE,
TV_CPROP1,
TV_PRE,
TV_HOIST,
TV_CPROP2,
TV_LSM,
TV_BYPASS,
TV_TRACER,
TV_WEB,
TV_CSE2,
TV_BRANCH_PROB,
TV_VPT,
TV_FLOW,
TV_COMBINE,
TV_IFCVT,
TV_REGMOVE,
TV_MODE_SWITCH,
TV_SMS,
TV_SCHED,
TV_LOCAL_ALLOC,
TV_GLOBAL_ALLOC,
TV_RELOAD_CSE_REGS,
TV_GCSE_AFTER_RELOAD,
TV_FLOW2,
TV_IFCVT2,
TV_PEEPHOLE2,
TV_RENAME_REGISTERS,
TV_SCHED2,
TV_MACH_DEP,
TV_DBR_SCHED,
TV_REORDER_BLOCKS,
TV_SHORTEN_BRANCH,
TV_REG_STACK,
TV_FINAL,
TV_SYMOUT,
TV_VAR_TRACKING,
TV_REST_OF_COMPILATION,
  TIMEVAR_LAST
}
timevar_id_t;
