#include "introspector_forwards.h"
#include "introspector_stdlib.h"
#include "introspector_gcc.h"
#include "introspector_rdf.h"
#include "introspector_inline.h"
#include "introspector_pool.h"

#include "introspector.h"

//#define MAXNODES 1000000
//#define MAXSTMTS 400000
//#define MAXURIS  1000
#include <string.h>
#include <stdlib.h>

static int maxwidth=0;
//static librdf_node node[MAXNODES];
//static librdf_statement statements[MAXSTMTS];
static librdf_model* model;
static librdf_serializer* serializer;
static librdf_storage* storage;
static librdf_world* world;

static long int nodecount=0;
static long int statementcount=0;
static char debug=0;

librdf_node* librdf_new_node(const unsigned char *string);

int librdf_model_print_statement(librdf_model* model, librdf_statement* statement){

  if (debug)
    printf ("STMT(%4ld)\t",	  statement->id);
  if (debug)
    printf ("%p[%ld]->%p",
	    (void*)statement->subject,
	    statement->subject->id,
	    (void*)statement->subject->str);
  if (debug)
    printf ("'%10s'\t",statement->subject->str);

  if (debug)
    printf ("%p[%ld]->%p",
	    (void*)&statement->predicate->name,
	    statement->predicate->name.id,
	    (void*)statement->predicate->name.str
	    );  
  if (debug)
    printf ("'%15s'\t",statement->predicate->name.str);
  if (debug)
    printf ("t%p[%ld]->%p",
	    (void*)statement->object,
	    statement->object->id,
	    (void*)statement->object->str
	    );
  if (debug)
    printf ("'%s'\n",statement->object->str);
}

int librdf_model_add_statement(librdf_model* model, librdf_statement* statement){
  model=0;

  introspector_predicate_add_object(
				    statement->predicate,
				    statement->object);
  introspector_predicate_add_subject(
				    statement->predicate,
				    statement->subject);

  // handle the statement
  librdf_model_print_statement(model, statement);
  introspector_dir_model_add_statement(model,statement);

  return 1;
}
void librdf_statement_set_subject(librdf_statement *statement, librdf_node *subject)
{
  if (subject)
    {
        if (debug)
	  printf("\tSubject:%p::%p->'%s'\n",statement,subject,subject->str);
    }
  else
    {
        if (debug)
	  printf("\tSub:'%p'\n",subject);
    }
  statement->subject=subject;
}

librdf_statement* librdf_new_statement(librdf_world* world){
  world=0;
  if (debug)
    printf ("New Statement %ld\n",statementcount);
  librdf_statement* statement=alloc_statement();
  statement->subject=0;
  statement->object=0;
  statement->predicate=0;
  statement->id=statementcount++;
  return statement;
}

int librdf_serializer_serialize_model(librdf_serializer* serializer, FILE *handle, librdf_uri* base_uri, librdf_model* model){
  serializer=0;
  handle=0;
  base_uri=0;
  model=0;
  return 1;
}

librdf_node* librdf_new_node(const unsigned char *string)
{
  librdf_node* ret;
  int width=strlen(string);
  maxwidth= width>maxwidth ? width:maxwidth;
  ret=alloc_node();

  if (width>NODE_WIDTH)
    {
      if (debug)
	printf("Expand string to %d",width);
      strncpy(ret->str,string,NODE_WIDTH);
    }
  else
    {
      strcpy(ret->str,string);
    }

  if (debug)
    printf ("\tLiteral %p->'%s' %ld\n",ret,string,nodecount);
  ret->id=nodecount;
  nodecount++;
  return ret;
}

librdf_node* librdf_new_node_from_literal(librdf_world* world, const unsigned char *string, const char *xml_language, int is_wf_xml)
{
  world=0;
  xml_language=0;
  is_wf_xml=0;
  return librdf_new_node(string);
}

librdf_node* librdf_new_node_from_uri_local_name(librdf_world* world, librdf_uri *uri, const unsigned char *local_name)
{
  world=0;
  uri=0;
  return librdf_new_node(local_name);
}

librdf_uri* librdf_new_uri (librdf_world *world, const unsigned char *string){
  return librdf_new_node(string);
}

void librdf_statement_set_object(librdf_statement *statement, librdf_node *object)
{
  if (debug)
    printf("\tObj:'%s'\n",object->str);
  statement->object=object;


}

void librdf_statement_set_predicate(librdf_statement *statement, librdf_node *predicate)
{
  if (debug)
    printf("\tPred:'%s'\n",predicate->str);
  statement->predicate=predicate;
}


void librdf_world_open(librdf_world *world){world=0;}

librdf_model* librdf_new_model(librdf_world *world, librdf_storage *storage, char* options_string){
  world=0;
  storage=0;
  options_string=0;
  return model;}

librdf_node* librdf_statement_get_object(librdf_statement *statement){statement=0;return 0;}

librdf_node* librdf_statement_get_predicate(librdf_statement *statement){
  statement=0;
  return 0;
}

librdf_serializer* librdf_new_serializer(librdf_world* world, const char *name, const char *mime_type, librdf_uri *type_uri){
  world=0;name=0;mime_type=0;type_uri=0;
return serializer;}

librdf_storage* librdf_new_storage(librdf_world *world, char *storage_name, char *name, char *options_string){
  world=0;storage_name=0;name=0;options_string=0;
return storage;}


void librdf_free_model(librdf_model *model){  model=0;}

void librdf_free_node(librdf_node *r){r=0;}

void librdf_free_serializer(librdf_serializer *serializer){  serializer=0;}

void librdf_free_storage(librdf_storage *storage){storage=0;}

void librdf_free_uri(librdf_uri *uri){uri=0;}

librdf_world* librdf_new_world(void){
  introspector_alloc();
  introspector_create_predicate_fields();
  introspector_node_types_init();
  return world;
}

void librdf_free_world(librdf_world *world){
  world=0;
  if (debug)
    printf("MaxWidth %d\n",maxwidth);
  introspector_node_types_delete();
  introspector_delete_predicate_fields ();
  introspector_free();

}
