#! /usr/bin/perl -w

# usage example
# perl /home/mdupont/gcc/distribute/gcc/gcc/introspector/process.pl OBJ/id-1 decl 1 | tee log.txt 
use strict;
use warnings;
use File::Spec;
use File::Path;
require Cwd;

use FindBin ();
use lib "$FindBin::Bin";
use Find ();
my  $debug_filter=0;


my $cwd = Cwd::getcwd();
my $start= shift @ARGV;
my $set= shift @ARGV;
my $create= shift @ARGV ||0;

# Set the variable $File::Find::dont_use_nlink if you're using AFS,
# since AFS cheats.

# for the convenience of &wanted calls, including -eval statements:
use vars qw/*name *dir *prune/;
*name   = *Introspector::Find::name;
*dir    = *Introspector::Find::dir;
*prune  = *Introspector::Find::prune;

sub wanted;


our %children_attrs=(
	       tree_code => "attribute",
	       modifier  => "attribute",
		     strg  => "attribute",
		     string  => "attribute",
	       linenumber  => "attribute",
	       filename  => "print",
	       used  => "attribute",
	       chan      => "list",
	       purp      => "list",
	       valu      => "list",
);

our %sets =
    (
     all => {
	 args  => "args",
	 argt  => "type",
	 body      => "child",
	 child_statement      => "child",
	 fn      => "child",
	 high=> "size",
	 low=> "size",
	 max=> "size",
	 min=> "size",
	 name      => "child",
	 name => "child",
	 op_0      => "child",
	 op_1      => "child",
	 op_2      => "child",
	 prec      => "size",
	 prms      => "child",
	 prms  => "args",
	 retn      => "type",
	 size => "size",
	 type  => "type",
	 unql      => "type",
	 vars      => "something",
	 algn => "type",
     },
     decl=>	     {
	 name      => "child",
	 args      => "child",
	 body      => "child",
	 op_0      => "child",
	 op_1      => "child",
	 op_2      => "child",
	 fn      => "child",
	 prms      => "child",
	 child_statement      => "child",
#	 vars      => "something",
     },
    types =>{
	name      => "child",
	algn => "type",
	argt => "type",
	args => "args",
	prms => "args",
	ptr  => "ptr",
	type=> "type",
	unql      => "type",
	retn      => "type",
    }
);

our %skip_children=(
		    scpe      => "skip",
		    size => "size",
		    min=> "size",
		    max=> "size",
		    high=> "size",
		    low=> "size",
		    prec      => "size",
	   );


our %follow_children;

sub prepare
{
    foreach my $k (keys %children_attrs)
    {
	$follow_children{$k}=$children_attrs{$k};
    }

    foreach my $k (keys %{$sets{$set}})
    {
	$follow_children{$k}=$sets{$set}{$k};
    }
}

prepare;



sub filter_children
{
    my $FN=shift;
    my $loc_pref=shift;
    my $rel = File::Spec->abs2rel( $loc_pref, $cwd ) ;

    if ($skip_children{$FN})
    {
#	warn "Skip $FN";
	warn "Skip $FN" if $debug_filter;;
    }
    elsif (my $type =$follow_children{$FN})
    {
	if ($type eq "print")
	{
	    warn "Found $FN" if $debug_filter;
	}

	warn "Going to follow $FN" if $debug_filter;;
    }
    elsif ($FN =~ /^id-/)
    {
		warn "Going to follow ID $FN" if $debug_filter;;
    }
    elsif (-l $loc_pref.$FN)
    {
	warn "Skip to follow LINK $FN" if $debug_filter;;
    }
    elsif ($rel  eq "nodes")
    {
	warn "NODES" if $debug_filter;;
    }
    else
    {
	warn "Skip OTHER $FN in $rel" if $debug_filter;;
	return 0;
    }
    return 1;
}

sub check_children
{
    my $p_dir =shift;

    warn "Start  $p_dir" if $debug_filter;;
    if ($p_dir=~ /\/OBJ$/)
    {
	# we are in the root
	return 1;
    }
    elsif ($p_dir=~ /\/OBJ\/id-\d+$/)
    {
#				  warn "Start  $p_dir" if $debug_filter;;
	return 1;
    }
    elsif ($p_dir=~ /\/scpe$/)
    {
	warn "scpe  $p_dir" if $debug_filter;;
	return 0;
    }


    elsif ($p_dir=~ /\/(\w+)$/)
    {
	if ($follow_children{$1})
	{
	    warn "Size  $p_dir" if $debug_filter;
	    return 1;
	}
	else
	{
	    return 0;
	}
    }
    else
    {
	warn "other $p_dir" if $debug_filter;;
	return 1;
    }
}

sub preprocess_children
{
    my @list=@_;
    my @ret;
    foreach my $x (@list)
    {
	if ($x =~ /^\./ )
	{
	}
	elsif ($x =~ /id-/ )
	{
	    push @ret,$x;
	}
	elsif ($follow_children{$x})
	{
	    warn "follow :$x:" if $debug_filter;;
	    push @ret,$x;
	}
	else
	{
	    warn "skipping :$x:" ;
	    warn "skipping :$x:" if $debug_filter;;
	}
	#warn "tEST:$_";
    }
    return @ret;
}


# Traverse desired filesystems
Introspector::Find::find(
		  {
		      wanted => \&wanted,
#		      follow => 1,
		      follow_fast=>1,
		      follow_skip=>2,
		      check_children => \&check_children,
		      skip_children => \%skip_children,
		      follow_children => \%follow_children,
		      filter_children => \&filter_children,

		      preprocess => \&preprocess_children,
		  },
			  $start
    );
# id-1
exit;


sub shortenpath
{
    my $name=shift;

    if ($name !~ /identifier_node/)
    {
	$name =~ s/id-(\d+)\/tree_code\/(\w+)/id-$1\/tc_$2\/$1/g; # replace a chan with c
    }
    else
    {
	return "";
    }

    $name =~ s/\/chan\/id-\d+/\/c\//g; # replace a chan with c
    $name =~ s/\/used\/\d+/\/used\//g; # used
    $name =~ s/\/filename\/filename:/fn:/g; # filename
    $name =~ s/\/modifier\//mod:/g; # modifier

    #body/child_statement/id-648
    foreach my $f qw(
		     tree_code
		     modifier
		     strg
		     filename
		     chan
		     used
		     purp
		     valu)
    {
	if ($name =~ /\/$f\/?$/)
	{
	    return "";
	}
    }

    foreach my $f qw(
			args
			argt
			body
			fn
			high
			low
			max
			min
			name
			name
			op_0
			op_1
			op_2
			prec
			prms
			prms
			retn
			size
			type
			unql
			vars
			algn
			)
    {
	$name =~ s/\/$f\/id-\d+/\/$f\//g; # replace the field/id-X with just the field

	if ($name =~ /\/$f\/?$/)
	{
	    return "";
	}
    }

    $name =~ s/\/valu\/id-\d+/\/v\//g; # replace the field/id-X with just the field

    foreach my $f qw(
		     args
		     argt
		     body
		     fn
		     high
		     low
		     max
		     min
		     name
		     name
		     op_0
		     op_1
		     op_2
		     prec
		     prms
		     prms
		     retn
		     size
		     type
		     unql
		     vars
		     algn
		     scpe

			)#		     valu
    {
	$name =~ s/\/$f\/id-\d+/\/$f\//g; # replace the field/id-X with just the field

	if ($name =~ /\/$f\/?$/)
	{
	    return "";
	}
    }

    $name =~s/\s+/:/g; #replace  spaces with +
    $name =~s/\/\//\//g; #replace  // with /
    $name =~ s/\/name\/strg\//\/names:/g; # namestring


##OBJ/id-1/body/vars/c/c_366/name/tc_identifier_nod


    $name =~ s/\/body\/child_statement\/id-(\d+)/\/bc_$1/g; # replace the body/child with something nicer

    $name =~ s/\/child_statement\/id-(\d+)/\/cs_$1/g; # replace the body/child with something nicer

    # now flatten the c/c intro acounted chain
    my @parts = split "/",$name;

    my $l="";
    my $c=0;
    my  @res;
    foreach my $p (@parts)
    {
	if ($p ne $l)
	{

	    if  ($c >0)
	    {
		push @res,"${l}_${c}";
	    }
	    else
	    {
#		push @res,"$l";
	    }

	    push @res,"$p";

	    $l = $p; # save
	    $c =0;
	}
	else
	{
	    $c++;
	}

    }

    return join "/",@res;
}

sub wanted {
    my ($dev,$ino,$mode,$nlink,$uid,$gid);


    if ((($dev,$ino,$mode,$nlink,$uid,$gid) = stat($_)))
    {
#	if($name =~  /strg\//  )
	{
#	    print("NEW_$name\n");
	    my $newname = shortenpath($name);
	    if ($newname)
	    {

		my $newpath="${cwd}/NEW/$newname";
		my $oldpath="${cwd}/$name";
		mkpath ("${cwd}/NEW/$newname") if $create;

		$oldpath =~ s/\/[^\/]+$//; # remove the last part of the old file!
		$oldpath =~ s/\/[^\/]+$//; # remove the last part of the old file! again. 
		symlink $oldpath,$newpath . "/l"; # 

		print "$newname -->\n\t$oldpath\n";


	    }
	}
    }
}

